"""
Creates an output .hpp file from a .dat file created by parsing a header
"""

import inspect
import pathlib
import pickle
import sys

from ..autowrap.context import HeaderContext, ClassContext
from ..autowrap.render_cls_trampoline_hpp import render_cls_trampoline_hpp


def _get_classes(hctx: HeaderContext):
    def _get_child_classes(c: ClassContext):
        for c in c.child_classes:
            yield c
            _get_child_classes(c)

    for cls in hctx.classes:
        yield cls
        yield from _get_child_classes(cls)


def _write_wrapper_cpp(input_dat: pathlib.Path, yml_id: str, output_hpp: pathlib.Path):
    with open(input_dat, "rb") as fp:
        hctx = pickle.load(fp)

    assert isinstance(hctx, HeaderContext)

    avail = []
    for cls in _get_classes(hctx):
        avail.append(cls.yml_id)
        if cls.yml_id == yml_id:
            break
    else:
        msg = [
            f"cannot find {yml_id} in {hctx.rel_fname}",
            f"- config: {hctx.orig_yaml}",
        ]

        if avail:
            msg.append("- found " + ", ".join(avail))

        if hctx.ignored_classes:
            msg.append("- ignored " + ", ".join(hctx.ignored_classes))

        raise ValueError("\n".join(msg))

    content = render_cls_trampoline_hpp(hctx, cls)
    output_hpp.write_text(content, encoding="utf-8")


def main():
    try:
        _, input_dat, yml_id, output_hpp = sys.argv
    except ValueError:
        print(inspect.cleandoc(__doc__ or ""), file=sys.stderr)
        sys.exit(1)

    _write_wrapper_cpp(pathlib.Path(input_dat), yml_id, pathlib.Path(output_hpp))


if __name__ == "__main__":
    main()



import argparse
import sys

from .build_dep import BuildDep
from .update_yaml import YamlUpdater
from .create_imports import ImportCreator, UpdateInit
from .scan_headers import HeaderScanner


def main():
    parser = argparse.ArgumentParser(prog="semiwrap")
    parent_parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    for cls in (
        BuildDep,
        YamlUpdater,
        HeaderScanner,
        ImportCreator,
        UpdateInit,
    ):
        cls.add_subparser(parent_parser, subparsers).set_defaults(cls=cls)

    args = parser.parse_args()
    cmd = args.cls()
    retval = cmd.run(args)

    if retval is False:
        retval = 1
    elif retval is True:
        retval = 0
    elif isinstance(retval, int):
        pass
    else:
        retval = 0

    sys.exit(retval)


if __name__ == "__main__":
    main()



from __future__ import annotations

import collections
import dataclasses
import os
import pathlib
import pprint
import sys
import sysconfig
import typing as T

from .casters import PKGCONF_CASTER_EXT
from .config.autowrap_yml import AutowrapConfigYaml
from .config.pyproject_toml import ExtensionModuleConfig, TypeCasterConfig
from .pkgconf_cache import PkgconfCache
from .pyproject import PyProject
from .util import relpath_walk_up

import toposort


@dataclasses.dataclass
class Entrypoint:
    """
    Represents a python entrypoint that needs to be created when the wheel is installed

    .. seealso:: https://packaging.python.org/en/latest/specifications/entry-points/
    """

    name: str
    group: str
    package: str


@dataclasses.dataclass(frozen=True)
class CppMacroValue:
    """
    Represents the value of a macro defined by the C++ compiler
    """

    name: str


@dataclasses.dataclass(frozen=True)
class CompilerInfo:
    """
    Represents N arguments to be injected:

    - argument_syntax: gcc, msvc, or other
    - c++ standard
    - actual arguments to run the compiler

    """

    pass


@dataclasses.dataclass(frozen=True)
class LocalDependency:
    name: str
    include_paths: T.Tuple[pathlib.Path, ...]
    depends: T.Tuple[T.Union[LocalDependency, str], ...]


@dataclasses.dataclass(frozen=True)
class InputFile:
    #: this path is always relative to the project root
    path: pathlib.Path


@dataclasses.dataclass(frozen=True)
class OutputFile:
    name: str


@dataclasses.dataclass(frozen=True)
class Depfile:
    # this is needed anytime a command generates something based on something
    # that is more than just the input file
    name: str


@dataclasses.dataclass(frozen=True)
class BuildTargetOutput:
    # references a specific build target output
    target: BuildTarget
    output_index: int


@dataclasses.dataclass(frozen=True)
class BuildTarget:
    command: str

    args: T.Tuple[
        T.Union[
            str,
            pathlib.Path,
            InputFile,
            OutputFile,
            Depfile,
            BuildTarget,
            BuildTargetOutput,
            CppMacroValue,
            CompilerInfo,
        ],
        ...,
    ]

    # Install path is always relative to py.get_install_dir(pure: false)
    install_path: T.Optional[pathlib.Path]


@dataclasses.dataclass(frozen=True)
class ExtensionModule:

    #: variable name/prefix in the build file
    name: str

    #: full package name of installed extension
    package_name: str

    sources: T.Tuple[BuildTarget, ...]
    depends: T.Tuple[T.Union[LocalDependency, str], ...]

    # extra include directories that won't be found via depends
    include_directories: T.Tuple[pathlib.Path, ...]

    defines: T.Tuple[T.Tuple[str, T.Union[str, int]], ...]

    # Install path is always relative to py.get_install_dir(pure: false)
    install_path: pathlib.Path


class PlanError(Exception):
    pass


def _split_ns(name: str) -> T.Tuple[str, str]:
    ns = ""
    idx = name.rfind("::")
    if idx != -1:
        ns = name[:idx]
        name = name[idx + 2 :]
    return ns, name


class _BuildPlanner:
    def __init__(self, project_root: pathlib.Path, missing_yaml_ok: bool = False):

        self.project_root = project_root
        self.missing_yaml_ok = missing_yaml_ok

        self.pyproject = PyProject(project_root / "pyproject.toml")
        self.pkgcache = PkgconfCache()
        self.pyproject_input = InputFile(pathlib.Path("pyproject.toml"))

        self.pyi_targets: T.List[BuildTarget] = []
        self.pyi_args = []

        self.local_caster_targets: T.Dict[str, BuildTargetOutput] = {}
        self.local_dependencies: T.Dict[str, LocalDependency] = {}

        sw_path = self.pkgcache.get("semiwrap").type_casters_path
        assert sw_path is not None
        self.semiwrap_type_caster_path = sw_path

    def generate(self):

        projectcfg = self.pyproject.project

        #
        # Export type casters
        # .. this probably should be its own hatchling plugin?
        #

        for name, caster_cfg in projectcfg.export_type_casters.items():
            yield from self._process_export_type_caster(name, caster_cfg)

        # This is needed elsewhere
        self._cpp_macro = CppMacroValue("__cplusplus")
        yield self._cpp_macro

        #
        # Generate extension modules
        #

        for package_name, extension in self._sorted_extension_modules():
            try:
                yield from self._process_extension_module(package_name, extension)
            except Exception as e:
                raise PlanError(f"{package_name} failed") from e

        # TODO: this conditional probably should be done in the build system instead
        # cannot build pyi files when cross-compiling
        if not (
            "_PYTHON_HOST_PLATFORM" in os.environ
            or "PYTHON_CROSSENV" in os.environ
            or os.environ.get("SEMIWRAP_SKIP_PYI") == "1"
        ):
            # Make a pyi for every module
            # - they depend on every module because it needs a working environment
            #   and the user might import something
            # - if there's a subpkg this fails, need to think about it
            for pyi_target in self.pyi_targets:
                yield BuildTarget(
                    command="make-pyi",
                    args=pyi_target.args + tuple(self.pyi_args),
                    install_path=pyi_target.install_path,
                )

    def _resolve_dep(self, dname: str):
        return self.local_dependencies.get(dname, dname)

    def _process_export_type_caster(self, name: str, caster_cfg: TypeCasterConfig):

        # Need to generate the data file and the .pc file
        caster_target = BuildTarget(
            command="publish-casters",
            args=(
                self.pyproject_input,
                name,
                OutputFile(f"{name}{PKGCONF_CASTER_EXT}"),
                OutputFile(f"{name}.pc"),
            ),
            install_path=pathlib.Path(*caster_cfg.pypackage.split(".")),
        )
        yield caster_target

        # Need an entrypoint to point at the .pc file
        yield Entrypoint(group="pkg_config", name=name, package=caster_cfg.pypackage)

        dep = self.pkgcache.add_local(
            name=name,
            includes=[self.project_root / inc for inc in caster_cfg.includedir],
            requires=caster_cfg.requires,
        )
        caster_dep = LocalDependency(
            name=dep.name,
            include_paths=tuple(dep.include_path),
            depends=tuple([self._resolve_dep(cd) for cd in caster_cfg.requires]),
        )
        self.local_dependencies[dep.name] = caster_dep
        yield caster_dep

        # The .pc file cannot be used in the build, but the data file must be, so
        # store it so it can be used elsewhere
        self.local_caster_targets[name] = BuildTargetOutput(caster_target, 0)

    def _sorted_extension_modules(
        self,
    ) -> T.Generator[T.Tuple[str, ExtensionModuleConfig], None, None]:
        # sort extension modules by dependencies, that way modules can depend on other modules
        # also declared in pyproject.toml without needing to worry about ordering in the file
        by_name = {}
        to_sort: T.Dict[str, T.Set[str]] = {}

        for package_name, extension in self.pyproject.project.extension_modules.items():
            if extension.ignore:
                continue

            name = extension.name or package_name.replace(".", "_")
            by_name[name] = (package_name, extension)

            deps = to_sort.setdefault(name, set())
            for dep in extension.wraps:
                deps.add(dep)
            for dep in extension.depends:
                deps.add(dep)

        for name in toposort.toposort_flatten(to_sort, sort=True):
            data = by_name.get(name)
            if data:
                yield data

    def _process_extension_module(
        self, package_name: str, extension: ExtensionModuleConfig
    ):
        package_path_elems = package_name.split(".")
        parent_package = ".".join(package_path_elems[:-1])
        module_name = package_path_elems[-1]
        package_path = pathlib.Path(*package_path_elems[:-1])
        varname = extension.name or package_name.replace(".", "_")

        # Detect the location of the package in the source tree
        package_init_py = self.pyproject.package_root / package_path / "__init__.py"
        self.pyi_args += [parent_package, package_init_py.as_posix()]

        depends = self.pyproject.get_extension_deps(extension)
        depends.append("semiwrap")

        # Search path for wrapping is dictated by package_path and wraps
        search_path, include_directories_uniq, caster_json_file, libinit_modules = (
            self._prepare_dependency_paths(depends, extension)
        )

        includes = [
            self.project_root / pathlib.PurePosixPath(inc) for inc in extension.includes
        ]
        search_path.extend(includes)
        for inc in includes:
            include_directories_uniq[inc] = True

        # Search the package path last
        search_path.append(self.pyproject.package_root / package_path)

        all_type_casters = BuildTarget(
            command="resolve-casters",
            args=(
                OutputFile(f"{varname}.casters.pkl"),
                Depfile(f"{varname}.casters.d"),
                *caster_json_file,
            ),
            install_path=None,
        )
        yield all_type_casters

        #
        # Generate init.py for loading dependencies
        #

        libinit_module = None
        if libinit_modules:
            libinit_py = extension.libinit or f"_init_{module_name}.py"
            libinit_tgt = BuildTarget(
                command="gen-libinit-py",
                args=(OutputFile(libinit_py), *libinit_modules),
                install_path=package_path,
            )

            libinit_module = f"{parent_package}.{libinit_py}"[:-3]
            self.pyi_args += [libinit_module, libinit_tgt]

            yield libinit_tgt

        #
        # Publish a .pc file for this module
        #

        pc_args = [
            package_name,
            varname,
            self.pyproject_input,
            OutputFile(f"{varname}.pc"),
        ]
        if libinit_module:
            pc_args += ["--libinit-py", libinit_module]

        yield BuildTarget(
            command="gen-pkgconf",
            args=tuple(pc_args),
            install_path=package_path,
        )

        yield Entrypoint(group="pkg_config", name=varname, package=parent_package)

        #
        # Process the headers
        #

        datfiles, module_sources, subpackages = yield from self._process_headers(
            extension,
            package_path,
            include_directories_uniq.keys(),
            search_path,
            all_type_casters,
        )

        modinit = BuildTarget(
            command="gen-modinit-hpp",
            args=(
                module_name,
                OutputFile(f"semiwrap_init.{package_name}.hpp"),
                *datfiles,
            ),
            install_path=None,
        )
        module_sources.append(modinit)
        yield modinit

        #
        # Emit the module
        #

        # Use a local dependency to store everything so it can be referenced elsewhere
        cached_dep = self.pkgcache.add_local(
            name=varname,
            includes=[*includes, self.pyproject.package_root / package_path],
            requires=depends,
            libinit_py=libinit_module,
        )
        local_dep = LocalDependency(
            name=cached_dep.name,
            include_paths=tuple(cached_dep.include_path),
            depends=tuple(self._resolve_dep(dep) for dep in depends),
        )
        yield local_dep
        self.local_dependencies[local_dep.name] = local_dep

        modobj = ExtensionModule(
            name=varname,
            package_name=package_name,
            sources=tuple(module_sources),
            depends=(local_dep,),
            include_directories=tuple(),
            defines=tuple(extension.defines.items()),
            install_path=package_path,
        )
        yield modobj

        self.pyi_args += [package_name, modobj]

        # This is not yielded here because pyi targets need to depend on all modules
        # via self.pyi_args.
        # - The output .pyi files vary based on whether there are subpackages or not. If no
        #   subpackage, it's {module_name}.pyi. If there are subpackages, it becomes
        #   {module_name}/__init__.pyi and {module_name}/{subpackage}.pyi
        # - As long as a user doesn't manually bind a subpackage our detection works here
        #   but if we need to allow that then will need to declare subpackages in pyproject.toml
        # .. this breaks if there are sub-sub packages, don't do that please

        base_pyi_elems = package_name.split(".")

        if subpackages:
            pyi_elems = base_pyi_elems + ["__init__.pyi"]
            pyi_args = [
                pathlib.PurePath(*pyi_elems).as_posix(),
                OutputFile("__init__.pyi"),
            ]
            for subpackage in subpackages:
                pyi_elems = base_pyi_elems + [f"{subpackage}.pyi"]
                pyi_args += [
                    pathlib.PurePath(*pyi_elems).as_posix(),
                    OutputFile(f"{subpackage}.pyi"),
                ]

            self.pyi_targets.append(
                BuildTarget(
                    command="make-pyi",
                    args=(package_name, *pyi_args, "--"),
                    install_path=package_path / module_name,
                )
            )

        else:
            base_pyi_elems[-1] = f"{base_pyi_elems[-1]}.pyi"

            self.pyi_targets.append(
                BuildTarget(
                    command="make-pyi",
                    args=(
                        package_name,
                        pathlib.PurePath(*base_pyi_elems).as_posix(),
                        OutputFile(f"{module_name}.pyi"),
                        "--",
                    ),
                    install_path=package_path,
                )
            )

    def _locate_type_caster_json(
        self,
        depname: str,
        caster_json_file: T.List[T.Union[BuildTargetOutput, pathlib.Path]],
    ):
        checked = set()
        to_check = collections.deque([depname])
        while to_check:
            name = to_check.popleft()
            checked.add(name)

            entry = self.pkgcache.get(name)

            if name in self.local_caster_targets:
                caster_json_file.append(self.local_caster_targets[name])
            else:
                tc = entry.type_casters_path
                if tc and tc not in caster_json_file:
                    caster_json_file.append(tc)

            for req in entry.requires:
                if req not in checked:
                    to_check.append(req)

    def _prepare_dependency_paths(
        self, depends: T.List[str], extension: ExtensionModuleConfig
    ):
        search_path: T.List[pathlib.Path] = []
        include_directories_uniq: T.Dict[pathlib.Path, bool] = {}
        caster_json_file: T.List[T.Union[BuildTargetOutput, pathlib.Path]] = []
        libinit_modules: T.List[str] = []

        # Add semiwrap default type casters
        caster_json_file.append(self.semiwrap_type_caster_path)

        for dep in depends:
            entry = self.pkgcache.get(dep)
            include_directories_uniq.update(
                dict.fromkeys(entry.full_include_path, True)
            )

            # extend the search path if the dependency is in 'wraps'
            if dep in extension.wraps:
                search_path.extend(entry.include_path)

            self._locate_type_caster_json(dep, caster_json_file)

            if entry.libinit_py:
                libinit_modules.append(entry.libinit_py)

        return search_path, include_directories_uniq, caster_json_file, libinit_modules

    def _process_headers(
        self,
        extension: ExtensionModuleConfig,
        package_path: pathlib.Path,
        include_directories_uniq: T.Iterable[pathlib.Path],
        search_path: T.List[pathlib.Path],
        all_type_casters: BuildTarget,
    ):
        datfiles: T.List[BuildTarget] = []
        module_sources: T.List[BuildTarget] = []
        subpackages: T.Set[str] = set()
        define_args = []

        yaml_path = self.pyproject.get_extension_yaml_path(extension)

        if extension.defines:
            for dname, dvalue in extension.defines.items():
                define_args += ["-D", f"{dname} {dvalue}"]

        for yml, hdr in self.pyproject.get_extension_headers(extension):
            yml_input = InputFile(yaml_path / f"{yml}.yml")

            try:
                ayml = AutowrapConfigYaml.from_file(self.project_root / yml_input.path)
            except FileNotFoundError:
                if not self.missing_yaml_ok:
                    msg = f"{self.project_root / yml_input.path}: use `python3 -m semiwrap update-yaml` to generate"
                    raise FileNotFoundError(msg) from None
                ayml = AutowrapConfigYaml()

            # find the source header
            h_input, h_root = self._locate_header(hdr, search_path)

            header2dat_args = []
            for inc in include_directories_uniq:
                header2dat_args += ["-I", inc]

            header2dat_args.extend(define_args)

            # https://github.com/pkgconf/pkgconf/issues/391
            header2dat_args += ["-I", sysconfig.get_path("include")]

            header2dat_args += ["--cpp", self._cpp_macro]

            header2dat_args.append(yml)
            header2dat_args.append(yml_input)
            header2dat_args.append(h_input)
            header2dat_args.append(h_root)
            header2dat_args.append(all_type_casters)
            header2dat_args.append(OutputFile(f"{yml}.dat"))
            header2dat_args.append(Depfile(f"{yml}.d"))
            header2dat_args.append(CompilerInfo())

            datfile = BuildTarget(
                command="header2dat", args=tuple(header2dat_args), install_path=None
            )
            yield datfile
            datfiles.append(datfile)

            # Every header has a .cpp file for binding
            cppfile = BuildTarget(
                command="dat2cpp",
                args=(datfile, OutputFile(f"{yml}.cpp")),
                install_path=None,
            )
            module_sources.append(cppfile)
            yield cppfile

            # Detect subpackages
            if ayml.defaults.subpackage:
                subpackages.add(ayml.defaults.subpackage)

            for f in ayml.functions.values():
                if f.ignore:
                    continue
                if f.subpackage:
                    subpackages.add(f.subpackage)
                for f in f.overloads.values():
                    if f.subpackage:
                        subpackages.add(f.subpackage)

            for e in ayml.enums.values():
                if e.ignore:
                    continue
                if e.subpackage:
                    subpackages.add(e.subpackage)

            # Every class gets a trampoline file, but some just have #error in them
            for name, ctx in ayml.classes.items():
                if ctx.ignore:
                    continue

                if ctx.subpackage:
                    subpackages.add(ctx.subpackage)

                cls_ns, cls_name = _split_ns(name)
                cls_ns = cls_ns.replace(":", "_")
                trampoline = BuildTarget(
                    command="dat2trampoline",
                    args=(datfile, name, OutputFile(f"{cls_ns}__{cls_name}.hpp")),
                    install_path=package_path / "trampolines",
                )
                module_sources.append(trampoline)
                yield trampoline

            # Even more files if there are templates
            if ayml.templates:

                # Every template instantiation gets a cpp file to lessen compiler
                # memory requirements
                for i, (name, tctx) in enumerate(ayml.templates.items(), start=1):
                    if tctx.subpackage:
                        subpackages.add(tctx.subpackage)

                    tmpl_cpp = BuildTarget(
                        command="dat2tmplcpp",
                        args=(datfile, name, OutputFile(f"{yml}_tmpl{i}.cpp")),
                        install_path=None,
                    )
                    module_sources.append(tmpl_cpp)
                    yield tmpl_cpp

                # All of which use this hpp file
                tmpl_hpp = BuildTarget(
                    command="dat2tmplhpp",
                    args=(datfile, OutputFile(f"{yml}_tmpl.hpp")),
                    install_path=None,
                )
                module_sources.append(tmpl_hpp)
                yield tmpl_hpp

        return datfiles, module_sources, subpackages

    def _locate_header(self, hdr: str, search_path: T.List[pathlib.Path]):
        phdr = pathlib.PurePosixPath(hdr)
        for p in search_path:
            h_path = p / phdr
            if h_path.exists():
                # We should return this as an InputFile, but inputs must be relative to the
                # project root, which may not be the case on windows. Incremental build should
                # still work, because the header is included in a depfile
                return h_path, p
        raise FileNotFoundError(
            f"cannot locate {phdr} in {', '.join(map(str, search_path))}"
        )


def makeplan(project_root: pathlib.Path, missing_yaml_ok: bool = False) -> T.Generator[
    T.Union[BuildTarget, Entrypoint, LocalDependency, ExtensionModule, CppMacroValue],
    None,
]:
    """
    Given the pyproject.toml configuration for a semiwrap project, reads the
    configuration and generates a series of commands that can be used to parse
    the input headers and generate the needed source code from them.
    """
    planner = _BuildPlanner(project_root, missing_yaml_ok)
    yield from planner.generate()


def main():
    # this command only exists for debugging purposes
    try:
        _, project_root = sys.argv
    except ValueError:
        print(f"{sys.argv[0]} project_root", file=sys.stderr)
        sys.exit(1)

    output = list(makeplan(pathlib.Path(project_root)))
    pprint.pprint(output, indent=1)


if __name__ == "__main__":
    main()



# TODO: rename this

import os
import pathlib
import sys
import traceback
import typing as T

from hatchling.plugin import hookimpl
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from .render_meson import render_meson_to_file

from .config.pyproject_toml import SemiwrapHatchlingConfig
from .config.util import parse_input


@hookimpl
def hatch_register_build_hook():
    return SemiwrapBuildHook


class SemiwrapBuildHook(BuildHookInterface):
    """
    Sets up code generation to be ran by meson
    """

    PLUGIN_NAME = "semiwrap"

    def initialize(self, version: str, build_data: T.Dict[str, T.Any]) -> None:
        # Only needed for building wheels
        if self.target_name != "wheel":
            return

        project_root = pathlib.Path(self.root).resolve()

        config = parse_input(
            self.config, SemiwrapHatchlingConfig, "[tool.hatch.build.hooks.semiwrap]"
        )
        stage0_build_path = project_root / config.autogen_build_path
        stage0_meson_build = stage0_build_path / "meson.build"
        stage0_gitignore = stage0_build_path / ".gitignore"

        if config.module_build_path is not None:
            stage1_build_path = project_root / config.module_build_path
        else:
            stage1_build_path = stage0_build_path / "modules"

        stage1_meson_build = stage1_build_path / "meson.build"

        # This is used to generate files installed into a directory called 'trampolines'
        # so due to https://github.com/mesonbuild/meson/issues/2320 this must also be
        # in a directory called 'trampolines'
        # - unlike stage0 and stage1, this is included by stage0
        trampoline_build_path = stage0_build_path / "trampolines"
        trampoline_meson_build = trampoline_build_path / "meson.build"

        try:
            eps = render_meson_to_file(
                project_root,
                stage0_meson_build,
                stage1_meson_build,
                trampoline_meson_build,
            )
        except Exception as e:
            # Reading the stack trace is annoying, most of the time the exception content
            # is enough to figure out what you did wrong.
            if os.environ.get("SEMIWRAP_ERROR_VERBOSE") == "1":
                raise

            msg = [
                "ERROR: exception occurred when processing `pyproject.toml`\n\n",
            ]

            msg += traceback.format_exception_only(type(e), e)
            cause = e.__context__
            while cause is not None:
                if "prepare_metadata_for_build_editable" in str(cause):
                    break

                el = traceback.format_exception_only(type(cause), cause)
                el[0] = f"- caused by {el[0]}"
                msg += el

                if cause.__suppress_context__:
                    break

                cause = cause.__context__

            msg.append(
                "\nSet environment variable SEMIWRAP_ERROR_VERBOSE=1 for stacktrace"
            )

            print("".join(msg), file=sys.stderr)
            sys.exit(1)

        if eps:
            # .. not documented but it works?
            for ep in eps:
                g = self.metadata.core.entry_points.setdefault(ep.group, {})
                g[ep.name] = ep.package

        if not stage0_gitignore.exists():
            stage0_gitignore.write_text(
                "/meson.build\n"
                f"/{stage1_build_path.name}/meson.build\n"
                f"/{trampoline_build_path.name}/meson.build\n"
            )



"""
Creates a per-module data file mapping type names to header files containing
pybind11 type caster implementations.
"""

import inspect
import pathlib
import pickle
import sys

from ..casters import CastersData, load_typecaster_json_data, TypeData
from ..depfile import Depfile


def _update_all_casters(type_caster_cfg: pathlib.Path, all_casters: CastersData):
    data = load_typecaster_json_data(type_caster_cfg)

    # flatten it
    for item in data.headers:
        header = pathlib.Path(item.header)
        if header.is_absolute():
            raise ValueError(
                f"{type_caster_cfg} contains absolute path to header: {header}"
            )

        for typ in item.types:
            td = TypeData(
                header=header, typename=typ, default_arg_cast=item.default_arg_cast
            )

            if typ not in all_casters:
                all_casters[typ] = td

            # in addition to the type, add a non-namespaced version too
            # - in theory this could cause conflicts, but in practice its useful
            # - this could be solved by making the parser resolve namespaces, but
            #   that has downsides too
            ntyp = typ.split("::")[-1]
            if ntyp not in all_casters:
                all_casters[ntyp] = td


def main():
    try:
        _, outfile_arg, depfile_arg = sys.argv[:3]
        caster_json_files = sys.argv[3:]
    except ValueError:
        print(inspect.cleandoc(__doc__ or ""), file=sys.stderr)
        sys.exit(1)

    outfile = pathlib.Path(outfile_arg)
    depfile = pathlib.Path(depfile_arg)

    d = Depfile(outfile)
    content: CastersData = {}

    for f in caster_json_files:
        path = pathlib.Path(f)
        d.add(path)
        _update_all_casters(path, content)

    # write the depfile
    d.write(depfile)

    # write the pickled data
    with open(outfile, "wb") as fp:
        pickle.dump(content, fp)


if __name__ == "__main__":
    main()



from ..config.autowrap_yml import (
    ClassData,
    EnumData,
    AutowrapConfigYaml,
    PropData,
    FunctionData,
    OverloadData,
)
from .context import OverloadTracker

from cxxheaderparser.types import Function

import dataclasses
import pathlib
from typing import Any, Dict, List, Optional, Tuple


@dataclasses.dataclass
class FnReportData:
    missing: bool = False
    overloads: Dict[str, bool] = dataclasses.field(default_factory=dict)
    tracker: OverloadTracker = dataclasses.field(default_factory=OverloadTracker)

    # need to be put into overloads if reports are being made
    deferred_signatures: List[Tuple[Function, bool]] = dataclasses.field(
        default_factory=list
    )


AttrMissingData = Dict[str, bool]
EnumMissingData = Dict[str, bool]
FnMissingData = Dict[str, FnReportData]


@dataclasses.dataclass
class ClsReportData:
    missing: bool
    attributes: AttrMissingData = dataclasses.field(default_factory=dict)
    enums: EnumMissingData = dataclasses.field(default_factory=dict)
    functions: FnMissingData = dataclasses.field(default_factory=dict)


def _merge_overload(data: FunctionData, overload: OverloadData) -> FunctionData:
    # merge overload information
    # - create a dictionary that contains things that haven't changed
    changes = {"overloads": {}}
    for f in dataclasses.fields(OverloadData):
        v = getattr(overload, f.name)
        if f.default_factory is not dataclasses.MISSING:
            default = f.default_factory()
        else:
            default = f.default

        if v != default:
            changes[f.name] = v

    return dataclasses.replace(data, **changes)


class GeneratorData:
    """
    Used by the hooks to retrieve user-specified generation data, and
    report to the user that there is data missing
    """

    data: AutowrapConfigYaml

    def __init__(self, data: AutowrapConfigYaml, data_fname: pathlib.Path):
        self.data = data
        self.data_fname = data_fname

        default_ignore = self.data.defaults.ignore
        self._default_enum_data = EnumData(ignore=default_ignore)
        self._default_fn_data = FunctionData(ignore=default_ignore)
        self._default_method_data = FunctionData()
        self._default_class_data = ClassData(ignore=default_ignore)
        self._default_class_enum_data = EnumData()

        # report data
        self.functions: FnMissingData = {}
        self.classes: Dict[str, ClsReportData] = {}
        self.enums: EnumMissingData = {}
        self.attributes: AttrMissingData = {}

    def get_class_data(self, name: str) -> Tuple[ClassData, bool]:
        """
        The 'name' is namespace::[parent_class::]class_name
        """
        data = self.data.classes.get(name)
        missing = data is None
        if missing:
            data = self._default_class_data

        self.classes[name] = ClsReportData(missing=missing)
        return data, missing

    def get_cls_enum_data(
        self, name: str, cls_key: str, cls_data: ClassData
    ) -> EnumData:
        if name is None:
            # TODO
            return self._default_class_enum_data
        data = cls_data.enums.get(name)
        if data is None:
            self.classes[cls_key].enums[name] = False
            data = self._default_class_enum_data

        return data

    def get_enum_data(self, name: str) -> EnumData:
        data = self.data.enums.get(name)
        if data is None:
            self.enums[name] = False
            data = self._default_enum_data
        return data

    def get_function_data(
        self,
        name: str,
        fn: Function,
        cls_key: Optional[str] = None,
        cls_data: Optional[ClassData] = None,
        is_private: bool = False,
    ) -> Tuple[FunctionData, OverloadTracker]:
        if cls_data and cls_key:
            data = cls_data.methods.get(name)
            report_base = self.classes[cls_key].functions
        else:
            data = self.data.functions.get(name)
            report_base = self.functions

        report_data = report_base.get(name)
        if not report_data:
            report_data = FnReportData()
            report_base[name] = report_data

        missing = data is None
        report_data.missing = missing and not is_private

        # When retrieving function data, we have to take into account which overload
        # is being processed, so that the user can customize each overload uniquely
        # if desired

        # most functions don't have overloads, so instead of computing the
        # signature each time we defer it until we actually need to use it

        if missing:
            if cls_key:
                data = self._default_method_data
            else:
                data = self._default_fn_data
            report_data.deferred_signatures.append((fn, is_private))
        elif not data.overloads:
            report_data.deferred_signatures.append((fn, is_private))
        else:
            # When there is overload data present, we have to actually compute
            # the signature of every function
            signature = self._get_function_signature(fn)
            overload = data.overloads.get(signature)
            missing = overload is None
            if not missing and overload:
                data = _merge_overload(data, overload)
            report_data.overloads[signature] = is_private or not missing

        report_data.tracker.add_overload()
        return data, report_data.tracker

    def add_using_decl(
        self, name: str, cls_key: str, cls_data: ClassData, is_private: bool
    ):
        # copied from get_function_data
        data = cls_data.methods.get(name)
        report_base = self.classes[cls_key].functions

        report_data = report_base.get(name)
        if not report_data:
            report_data = FnReportData()
            report_base[name] = report_data

        missing = data is None
        report_data.missing = missing and not is_private

        # We count this as an overload because it might be
        report_data.tracker.add_overload()

    def get_cls_prop_data(
        self, name: str, cls_key: str, cls_data: ClassData
    ) -> PropData:
        data = cls_data.attributes.get(name)
        if data is None:
            self.classes[cls_key].attributes[name] = False
            data = PropData()

        return data

    def get_prop_data(self, name) -> PropData:
        data = self.data.attributes.get(name)
        if data is None:
            self.attributes[name] = False
            data = PropData()

        return data

    def get_missing(self):
        """
        Generate a structure that can be copy/pasted into the generation
        data yaml and print it out if there's missing data
        """

        ignore_default = self.data.defaults.ignore
        report_missing = True
        if ignore_default and not self.data.defaults.report_ignored_missing:
            report_missing = False

        # note: sometimes we have strings from CppHeaderParser that aren't
        # strings, so we need to cast them to str so yaml doesn't complain

        data = self._process_missing(
            self.attributes,
            self.functions,
            self.enums,
            "functions",
            ignore_default,
            report_missing,
        )

        all_cls_data = {}
        for cls_key, cls_data in self.classes.items():
            if cls_data.missing and not report_missing:
                continue

            result = self._process_missing(
                cls_data.attributes,
                cls_data.functions,
                cls_data.enums,
                "methods",
                False,
                True,
            )
            if result or cls_data.missing:
                if ignore_default and cls_data.missing:
                    result["ignore"] = True
                all_cls_data[str(cls_key)] = result
        if all_cls_data:
            data["classes"] = all_cls_data

        return data

    def _process_missing(
        self,
        attrs: AttrMissingData,
        fns: FnMissingData,
        enums: EnumMissingData,
        fn_key: str,
        ignore_default: bool,
        report_missing: bool,
    ):
        data: Dict[str, Dict[str, Dict]] = {}

        # attributes
        if attrs:
            for y in attrs.keys():
                assert isinstance(y, str)

            data["attributes"] = {str(n): {} for n in attrs.keys()}

        # enums
        if enums:
            enums_report = {}
            for en, enum_present in enums.items():
                if not enum_present and not report_missing:
                    continue
                enums_report[en] = {}
                if ignore_default:
                    enums_report[en]["ignore"] = True
            if enums_report:
                data["enums"] = enums_report

        # functions
        fn_report = {}
        for fn, fndata in fns.items():
            if fndata.missing and not report_missing:
                continue

            fn = str(fn)
            overloads = fndata.overloads
            deferred_signatures = fndata.deferred_signatures
            overloads_count = len(overloads) + len(deferred_signatures)
            if overloads_count > 1:
                # process each deferred signature
                for dfn, v in deferred_signatures:
                    signature = self._get_function_signature(dfn)
                    overloads[signature] = v

                has_data = all(overloads.values())
            else:
                has_data = not fndata.missing

            if not has_data:
                d = {}
                if fn == "swap":
                    d = {"ignore": True}

                if overloads_count > 1:
                    fn_report[fn] = {
                        "overloads": {
                            k: dict(**d) for k, v in overloads.items() if not v
                        }
                    }

                    for k, v in fn_report[fn]["overloads"].items():
                        if "initializer_list" in k:
                            v["ignore"] = True
                        if ignore_default:
                            v["ignore"] = True
                else:
                    fn_report[fn] = d
                    if ignore_default:
                        d["ignore"] = True
        if fn_report:
            data[fn_key] = fn_report

        return data

    def _get_function_signature(self, fn: Function) -> str:
        """
        Only includes the names of parameters and a [const] indicator if needed
        """

        signature = ", ".join(
            f"{p.type.format()}..." if p.param_pack else p.type.format()
            for p in fn.parameters
        )

        if getattr(fn, "const", False):
            if signature:
                signature = f"{signature} [const]"
            else:
                signature = "[const]"

        # constexpr and non-constexpr cannot be overloaded, so don't include it
        # elif fn.constexpr:
        #     if signature:
        #         signature = f"{signature} [constexpr]"
        #     else:
        #         signature = "[constexpr]"

        return signature



"""
Usage: module_name output_hpp input_dat[, input_dat...]

Generates a header file that contains initialization functions for pybind11 bindings

You must include the header "autogen_module_init.hpp", and call initWrapper() from
your pybind11 module declaration.
"""

import pathlib
import pickle
import sys
import typing as T

import toposort

from ..autowrap.buffer import RenderBuffer
from ..autowrap.context import HeaderContext


def _write_wrapper_hpp(
    module_name: str, output_hpp: pathlib.Path, *input_dat: pathlib.Path
):
    # Need to ensure that wrapper initialization is called in base order
    # so we have to toposort it here based on the class hierarchy determined
    # while parsing the header
    types2name = {}
    types2deps = {}
    ordering = []

    for datfile in input_dat:
        with open(datfile, "rb") as fp:
            hctx = pickle.load(fp)

        assert isinstance(hctx, HeaderContext)

        name = hctx.hname
        dep = hctx.class_hierarchy

        # make sure objects without classes are also included!
        if not dep:
            ordering.append(name)

        for clsname, bases in dep.items():
            if clsname in types2name:
                raise ValueError(f"{name} ({input_dat}): duplicate class {clsname}")
            types2name[clsname] = name
            types2deps[clsname] = bases[:]

    to_sort: T.Dict[str, T.Set[str]] = {}
    for clsname, bases in types2deps.items():
        clsname = types2name[clsname]
        deps = to_sort.setdefault(clsname, set())
        for base in bases:
            base = types2name.get(base)
            if base and base != clsname:
                deps.add(base)

    ordering.extend(toposort.toposort_flatten(to_sort, sort=True))

    r = RenderBuffer()
    r.writeln("// This file is autogenerated, DO NOT EDIT")
    r.writeln("")
    r.write_trim(
        f"""
        #pragma once
        #include <semiwrap.h>

        // Use this to define your module instead of PYBIND11_MODULE
        #define SEMIWRAP_PYBIND11_MODULE(variable) PYBIND11_MODULE({module_name}, variable)

        // TODO: namespace semiwrap::autogen {{

        // forward declarations
        """
    )
    for name in ordering:
        r.write_trim(
            f"""
                void begin_init_{name}(py::module &m);
                void finish_init_{name}();
            """
        )
        r.writeln()

    r.writeln("static void initWrapper(py::module &m) {")
    with r.indent():
        for name in ordering:
            r.writeln(f"begin_init_{name}(m);")
        r.writeln()
        for name in ordering:
            r.writeln(f"finish_init_{name}();")

    r.writeln("}")

    output_hpp.write_text(r.getvalue(), encoding="utf-8")


def main():
    try:
        module_name = sys.argv[1]
        output_hpp = sys.argv[2]
        inputs = sys.argv[3:]
    except Exception as e:
        print(__doc__, file=sys.stderr)
        sys.exit(1)

    _write_wrapper_hpp(
        module_name, pathlib.Path(output_hpp), *[pathlib.Path(i) for i in inputs]
    )


if __name__ == "__main__":
    main()



import pathlib
import typing as T

from .pkgconf_cache import INITPY_VARNAME
from .util import relpath_walk_up


def make_pc_file(
    project_root: pathlib.Path,
    pcfile: pathlib.Path,
    pc_install_path: pathlib.Path,
    name: str,
    desc: str,
    version: str,
    includes: T.List[str],
    depends: T.List[str],
    libinit_py: T.Optional[str],
    generator_name: str,
):

    pc_content = [
        f"# automatically generated by {generator_name}",
        "prefix=${pcfiledir}",
    ]

    if libinit_py:
        pc_content.append(f"{INITPY_VARNAME}={libinit_py}")

    cflags = ["-I${prefix}"]

    for i, inc in enumerate(includes):
        includedir = project_root / pathlib.PurePosixPath(inc)
        rel = relpath_walk_up(includedir, pc_install_path)
        pc_content.append(f"inc{i}=${{prefix}}/{rel.as_posix()}")
        cflags.append(f"-I${{inc{i}}}")

    pc_content += [
        "",
        f"Name: {name}",
        f"Description: {desc}",
        f"Version: {version}",
        "Cflags: " + " ".join(cflags),
    ]

    if depends:
        requires = " ".join(depends)
        pc_content.append(f"Requires: {requires}")

    pc_content.append("")

    pcfile.write_text("\n".join(pc_content), encoding="utf-8")



import io
import contextlib
import inspect


class RenderBuffer:
    def __init__(self) -> None:
        self._buffer = io.StringIO()

        self._indent = ""
        self._indentlen = 0

    def getvalue(self) -> str:
        return self._buffer.getvalue()

    def rel_indent(self, spaces: int):
        self._indentlen += spaces
        self._indent = " " * self._indentlen

    @contextlib.contextmanager
    def indent(self, spaces: int = 2):
        self._indentlen += spaces
        self._indent = " " * self._indentlen
        try:
            yield
        finally:
            self._indentlen -= spaces
            self._indent = " " * self._indentlen

    def writeln(self, s: str = ""):
        if not s:
            self._buffer.write("\n")
        else:
            for line in s.splitlines(False):
                if line:
                    self._buffer.write(f"{self._indent}{line}\n")
                else:
                    self._buffer.write("\n")

    def write_trim(self, s: str):
        for line in inspect.cleandoc(s).splitlines(False):
            if line:
                self._buffer.write(f"{self._indent}{line}\n")
            else:
                self._buffer.write("\n")



from __future__ import annotations

import codecs
import os
import pathlib
import re
import sys
import typing as T

from .autowrap.buffer import RenderBuffer
from .makeplan import (
    Depfile,
    Entrypoint,
    InputFile,
    OutputFile,
    BuildTarget,
    BuildTargetOutput,
    ExtensionModule,
    LocalDependency,
    CppMacroValue,
    CompilerInfo,
    makeplan,
)
from .util import maybe_write_file, relpath_walk_up

# String escaping stolen from meson source code, Apache 2.0 license
# This is the regex for the supported escape sequences of a regular string
# literal, like 'abc\x00'
ESCAPE_SEQUENCE_SINGLE_RE = re.compile(
    r"""
    ( \\U[A-Fa-f0-9]{8}   # 8-digit hex escapes
    | \\u[A-Fa-f0-9]{4}   # 4-digit hex escapes
    | \\x[A-Fa-f0-9]{2}   # 2-digit hex escapes
    | \\[0-7]{1,3}        # Octal escapes
    | \\N\{[^}]+\}        # Unicode characters by name
    | \\[\\'abfnrtv]      # Single-character escapes
    )""",
    re.UNICODE | re.VERBOSE,
)


def _decode_match(match: T.Match[str]) -> str:
    return codecs.decode(match.group(0).encode(), "unicode_escape")


def _make_string(s: str):
    s = ESCAPE_SEQUENCE_SINGLE_RE.sub(_decode_match, s)
    return f"'{s}'"


VarTypes = T.Union[
    InputFile,
    OutputFile,
    BuildTarget,
    ExtensionModule,
    LocalDependency,
    CppMacroValue,
    CompilerInfo,
]


class VarCache:
    def __init__(self) -> None:
        self.cache: T.Dict[VarTypes, str] = {}

        # this might get annoying to debug, but for now this is easier...
        self.idx = 1

    def getvar(self, item: VarTypes) -> str:
        var = self.cache.get(item)
        if var is None:

            # TODO: probably should name variables according to their target type
            # to make it easier to debug

            if isinstance(item, InputFile):
                # .. this probably isn't right either, what should this be relative to?
                # .. TODO should use files()? but maybe only if this is actually a variable
                return _make_string(item.path.resolve().as_posix())
                # var = f"sw_in_{self.idx}"
            elif isinstance(item, OutputFile):
                return _make_string(item.name)
                # var = f"sw_out_{self.idx}"
            elif isinstance(item, BuildTarget):
                var = f"_sw_target_{self.idx}"
            elif isinstance(item, ExtensionModule):
                name = item.name.replace("-", "_")
                var = f"{name}_module"
            elif isinstance(item, LocalDependency):
                name = item.name.replace("-", "_")
                var = f"{name}_dep"
            elif isinstance(item, CppMacroValue):
                name = item.name
                var = f"_sw_cpp_var_{name}"
            elif isinstance(item, CompilerInfo):
                name = var = "_sw_compiler_info"
            else:
                assert False

            self.idx += 1
            self.cache[item] = var
        return var


def _render_build_target(r: RenderBuffer, vc: VarCache, bt: BuildTarget):
    outvar = vc.getvar(bt)

    bt_cmd = bt.command.replace("-", "_")
    cmd = [f"_sw_cmd_{bt_cmd}"]
    tinput = []
    toutput = []
    depfile = None

    for arg in bt.args:
        if isinstance(arg, str):
            cmd.append(_make_string(arg))
        elif isinstance(arg, pathlib.Path):
            cmd.append(_make_string(arg.resolve().as_posix()))
        elif isinstance(arg, (BuildTarget, InputFile)):
            cmd.append(f"'@INPUT{len(tinput)}@'")
            tinput.append(vc.getvar(arg))
        elif isinstance(arg, BuildTargetOutput):
            cmd.append(f"'@INPUT{len(tinput)}@'")
            tinput.append(f"{vc.getvar(arg.target)}[{arg.output_index}]")
        elif isinstance(arg, OutputFile):
            cmd.append(f"'@OUTPUT{len(toutput)}@'")
            toutput.append(vc.getvar(arg))
        elif isinstance(arg, Depfile):
            assert depfile is None, bt
            cmd.append("'@DEPFILE@'")
            depfile = _make_string(arg.name)
        elif isinstance(arg, ExtensionModule):
            cmd.append(f"'@INPUT{len(tinput)}@'")
            tinput.append(vc.getvar(arg))
        elif isinstance(arg, (CppMacroValue, CompilerInfo)):
            cmd.append(vc.getvar(arg))
        else:
            assert False, f"unexpected {arg!r} in {bt}"

    r.writeln(f"{outvar} = custom_target(")
    with r.indent(2):
        _render_meson_args(r, "command", cmd)
        if tinput:
            _render_meson_args(r, "input", tinput)
        if toutput:
            _render_meson_args(r, "output", toutput)

        if depfile:
            r.writeln(f"depfile: {depfile},")

        if bt.install_path is not None:
            install_path = _make_string(bt.install_path.as_posix())
            r.writeln(
                f"install_dir: sw_py.get_install_dir(pure: false) / {install_path},"
            )
            r.writeln("install: true,")

    r.writeln(")")


def _render_include_directories(
    r: RenderBuffer,
    incs: T.Sequence[pathlib.Path],
    meson_build_path: T.Optional[pathlib.Path],
):
    # meson wants these to be relative to meson.build
    # - only can do that if we're writing an output file
    if meson_build_path:
        meson_build_parent = meson_build_path.parent
        incs = [relpath_walk_up(p, meson_build_parent) for p in incs]

    _render_meson_args(
        r, "include_directories", [_make_string(inc.as_posix()) for inc in incs]
    )


def _render_meson_args(r: RenderBuffer, name: str, args: T.List[str]):
    r.writeln(f"{name}: [")
    with r.indent():
        for arg in args:
            r.writeln(f"{arg},")
    r.writeln("],")


def _render_module_stage0(
    r: RenderBuffer,
    vc: VarCache,
    m: ExtensionModule,
    meson_build_path: T.Optional[pathlib.Path],
):

    # variables generated here should be deterministic so that users can add
    # their own things to it, or use it directly

    r.writeln(f"# {m.package_name}")
    r.writeln(f"{m.name}_sources = []")
    r.writeln(f"{m.name}_deps = [declare_dependency(")
    with r.indent():

        if m.defines:
            r.writeln("compile_args: [")
            with r.indent():
                for dname, dvalue in m.defines:
                    r.writeln(_make_string(f"-D{dname}={dvalue}"))
            r.writeln("],")

        if m.sources:
            r.writeln("sources: [")
            with r.indent():
                for src in m.sources:
                    r.writeln(f"{vc.getvar(src)},")

            r.writeln("],")

        if m.depends:
            depnames = []
            for d in m.depends:
                if isinstance(d, LocalDependency):
                    depnames.append(vc.getvar(d))
                else:
                    depnames.append(f"dependency({_make_string(d)})")

            deps = ", ".join(depnames)
            r.writeln(f"dependencies: [{deps}],")

        if m.include_directories:
            _render_include_directories(r, m.include_directories, meson_build_path)

    r.writeln(")]")


def _render_module_stage1(
    r: RenderBuffer,
    vc: VarCache,
    m: ExtensionModule,
    meson_build_path: T.Optional[pathlib.Path],
):

    # variables generated here should be deterministic so that users can
    # use it directly if they wish

    subdir = _make_string(m.install_path.as_posix())
    module_name = _make_string(m.package_name.split(".")[-1])
    mvar = vc.getvar(m)

    r.writeln(f"# {m.package_name}")
    r.writeln(f"{mvar} = sw_py.extension_module(")
    with r.indent():
        r.write_trim(
            f"""
            {module_name},
            sources: [{m.name}_sources],
            dependencies: [{m.name}_deps],
            install: true,
            subdir: {subdir},        
        """
        )

        if m.include_directories:
            _render_include_directories(r, m.include_directories, meson_build_path)

    r.writeln(")")
    r.writeln()


def render_meson(
    project_root: pathlib.Path,
    stage0_path: T.Optional[pathlib.Path],
    stage1_path: T.Optional[pathlib.Path],
    trampolines_path: T.Optional[pathlib.Path],
) -> T.Tuple[str, str, str, T.List[Entrypoint]]:
    """
    Returns the contents of two meson.build files that build on each other, and
    any entry points that need to be created
    """

    eps: T.List[Entrypoint] = []

    r0 = RenderBuffer()
    r1 = RenderBuffer()
    t = RenderBuffer()
    vc = VarCache()

    # standard boilerplate here
    r0.write_trim(
        """
        # This file is automatically generated, DO NOT EDIT
        #
        # The generator's stable API includes variables that do not start with
        # an underscore. Any variables with an underscore may change in the future
        # without warning
        #
                 
        sw_py = import('python').find_installation()
        
        # internal commands for the autowrap machinery
        _sw_cmd_gen_libinit_py = [sw_py, '-m', 'semiwrap.cmd.gen_libinit']
        _sw_cmd_gen_pkgconf = [sw_py, '-m', 'semiwrap.cmd.gen_pkgconf']
        _sw_cmd_publish_casters = [sw_py, '-m', 'semiwrap.cmd.publish_casters']
        _sw_cmd_resolve_casters = [sw_py, '-m', 'semiwrap.cmd.resolve_casters']
        _sw_cmd_header2dat = [sw_py, '-m', 'semiwrap.cmd.header2dat']
        _sw_cmd_dat2cpp = [sw_py, '-m', 'semiwrap.cmd.dat2cpp']
        _sw_cmd_dat2trampoline = [sw_py, '-m', 'semiwrap.cmd.dat2trampoline']
        _sw_cmd_dat2tmplcpp = [sw_py, '-m', 'semiwrap.cmd.dat2tmplcpp']
        _sw_cmd_dat2tmplhpp = [sw_py, '-m', 'semiwrap.cmd.dat2tmplhpp']
        _sw_cmd_gen_modinit_hpp = [sw_py, '-m', 'semiwrap.cmd.gen_modinit_hpp']
        _sw_cmd_make_pyi = [sw_py, '-m', 'semiwrap.cmd.make_pyi']

        _sw_compiler_info = [
            meson.get_compiler('cpp').get_argument_syntax(),
            get_option('cpp_std'),
            meson.get_compiler('cpp').cmd_array()
        ]

        #
        # internal custom targets for generating wrappers
        #
    """
    )
    r0.writeln()

    # ... so, we could write out proper loops and stuff to make the output
    #     meson file easier to read, but I think expanded custom targets
    #     is simpler to generate

    plan = makeplan(pathlib.Path(project_root))
    macros: T.List[CppMacroValue] = []
    build_targets: T.List[BuildTarget] = []
    modules: T.List[ExtensionModule] = []
    pyi_targets: T.List[BuildTarget] = []
    local_deps: T.List[LocalDependency] = []
    trampoline_targets: T.List[BuildTarget] = []

    for item in plan:
        if isinstance(item, BuildTarget):
            if item.command == "make-pyi":
                # defer these to the end
                pyi_targets.append(item)
            elif item.command == "dat2trampoline":
                trampoline_targets.append(item)
            else:
                build_targets.append(item)
        elif isinstance(item, ExtensionModule):
            # defer these to the end
            modules.append(item)
        elif isinstance(item, Entrypoint):
            eps.append(item)
        elif isinstance(item, LocalDependency):
            local_deps.append(item)
        elif isinstance(item, CppMacroValue):
            macros.append(item)
        else:
            assert False

    if macros:
        for macro in macros:
            macro_name = _make_string(macro.name)
            r0.writeln(
                f"{vc.getvar(macro)} = meson.get_compiler('cpp').get_define({macro_name})"
            )
        r0.writeln()

    for target in build_targets:
        _render_build_target(r0, vc, target)

    if local_deps:
        r0.writeln()
        r0.write_trim(
            """
            #
            # Local dependencies
            #
        """
        )
        for ld in local_deps:
            r0.writeln(f"{vc.getvar(ld)} = declare_dependency(")
            with r0.indent():
                if ld.depends:
                    deps = []
                    for dep in ld.depends:
                        if isinstance(dep, LocalDependency):
                            deps.append(vc.getvar(dep))
                        else:
                            deps.append(f"dependency({_make_string(dep)})")

                    depstrs = ", ".join(deps)
                    r0.writeln(f"dependencies: [{depstrs}],")

                if ld.include_paths:
                    _render_include_directories(r0, ld.include_paths, stage0_path)

            r0.writeln(")")

    if trampoline_targets:
        t.writeln("# This file is automatically generated, DO NOT EDIT\n\n")
        for target in trampoline_targets:
            _render_build_target(t, vc, target)

        r0.writeln()
        r0.writeln("subdir('trampolines')")

    if modules:
        r0.writeln()
        r0.write_trim(
            """
            #
            # Module configurations
            #
        """
        )
        r0.writeln()

        r1.writeln("# This file is automatically generated, DO NOT EDIT\n\n")

        for module in modules:
            _render_module_stage0(r0, vc, module, stage0_path)
            _render_module_stage1(r1, vc, module, stage1_path)

        # TODO: this conditional probably should be done in meson instead
        # cannot build pyi files when cross-compiling
        if not (
            "_PYTHON_HOST_PLATFORM" in os.environ
            or "PYTHON_CROSSENV" in os.environ
            or os.environ.get("SEMIWRAP_SKIP_PYI") == "1"
        ):
            for target in pyi_targets:
                _render_build_target(r1, vc, target)

    return r0.getvalue(), r1.getvalue(), t.getvalue(), eps


def render_meson_to_file(
    project_root: pathlib.Path,
    stage0: pathlib.Path,
    stage1: pathlib.Path,
    trampolines: pathlib.Path,
) -> T.List[Entrypoint]:

    # because of https://github.com/mesonbuild/meson/issues/2320
    assert trampolines.parent.parent == stage0.parent

    s0_content, s1_content, t_content, eps = render_meson(
        pathlib.Path(project_root), stage0, stage1, trampolines
    )

    maybe_write_file(stage0, s0_content, encoding="utf-8")
    maybe_write_file(stage1, s1_content, encoding="utf-8")
    maybe_write_file(trampolines, t_content, encoding="utf-8")

    return eps


def main():
    def _usage() -> T.NoReturn:
        print(f"{sys.argv[0]} project_root stage0 [stage1]", file=sys.stderr)
        sys.exit(1)

    # this entrypoint only for debugging
    try:
        _, project_root = sys.argv
    except ValueError:
        _usage()

    s0_content, s1_content, t_content, _ = render_meson(
        pathlib.Path(project_root), None, None, None
    )

    print(s0_content)
    print("\n---\n")
    print(s1_content)
    print("\n---\n")
    print(t_content)


if __name__ == "__main__":
    main()



import dataclasses
import json
import pathlib
import typing as T

from .config.util import parse_input

PKGCONF_CASTER_EXT = ".pybind11.json"

#
# JSON caster data
#


@dataclasses.dataclass
class TypeCasterJsonHeader:
    header: str
    types: T.List[str]
    default_arg_cast: bool = False


#: content of .pybind11.json
@dataclasses.dataclass
class TypeCasterJsonData:
    """
    Stored in *.pybind11.json
    """

    headers: T.List[TypeCasterJsonHeader] = dataclasses.field(default_factory=list)


def load_typecaster_json_data(fname) -> TypeCasterJsonData:
    with open(fname, encoding="utf-8") as fp:
        return parse_input(json.load(fp), TypeCasterJsonData, fname)


def save_typecaster_json_data(fname: pathlib.Path, data: TypeCasterJsonData):
    with open(fname, "w", encoding="utf-8") as fp:
        json.dump(dataclasses.asdict(data), fp)


#
# Pickled caster lookup as stored by resolve_casters
#


@dataclasses.dataclass
class TypeData:
    header: pathlib.Path
    typename: str
    default_arg_cast: bool


#: content of pickle file used internally
CastersData = T.Dict[str, TypeData]



"""
Generates a file that loads shared library dependencies for the module

Arguments: pyproject.toml package.name /path/to/_init_name.py
"""

from __future__ import annotations

import inspect
import pathlib
import sys
import typing as T

from ..autowrap.buffer import RenderBuffer


def _write_libinit_py(
    init_py: pathlib.Path,
    modules: T.List[str],
) -> None:
    """
    :param init_py: the _init module for the library(ies) that is written out
    :param modules: the python modules to import
    """

    r = RenderBuffer()

    r.writeln("# This file is automatically generated, DO NOT EDIT")
    r.writeln("# fmt: off")
    r.writeln()

    for mod in modules:
        r.writeln(f"import {mod}")

    init_py.write_text(r.getvalue(), encoding="utf-8")


def main():
    try:
        _, libinit_py = sys.argv[:2]
        modules = sys.argv[2:]
    except ValueError:
        print(inspect.cleandoc(__doc__ or ""), file=sys.stderr)
        sys.exit(1)

    _write_libinit_py(
        pathlib.Path(libinit_py),
        modules,
    )


if __name__ == "__main__":
    main()



import os.path
import pathlib
import typing as T


def maybe_write_file(
    path: pathlib.Path, content: str, *, encoding: T.Optional[str] = None
) -> bool:
    # returns True if new content written
    if path.exists():
        with open(path, encoding=encoding) as fp:
            oldcontent = fp.read()
        if oldcontent == content:
            return False
    elif not path.parent.exists():
        path.parent.mkdir(parents=True)

    with open(path, "w", encoding=encoding) as fp:
        fp.write(content)

    return True


def relpath_walk_up(p: pathlib.Path, other: pathlib.Path) -> pathlib.Path:
    # walk_up=True was introduced in Python 3.12 so can't use that
    #   p.relative_to(other, walk_up=True)
    return pathlib.Path(os.path.relpath(p, other))



"""
Parses a header file and writes an intermediate dat file that other tools
can turn into other things
"""

import argparse
import pathlib
import pickle
import typing

from cxxheaderparser.options import ParserOptions
from cxxheaderparser import preprocessor

import yaml

from ..autowrap.cxxparser import parse_header
from ..autowrap.generator_data import GeneratorData
from ..casters import CastersData
from ..config.autowrap_yml import AutowrapConfigYaml


def format_missing(report) -> str:
    return (
        yaml.safe_dump(report, sort_keys=False)
        .replace(" {}", "")
        .replace("? ''\n          :", '"":')
        .replace("? ''\n      :", '"":')
    )


def generate_wrapper(
    *,
    name: str,
    src_yml: pathlib.Path,
    src_h: pathlib.Path,
    src_h_root: pathlib.Path,
    include_paths: typing.List[pathlib.Path],
    compiler_flavor: str,
    compiler_args: typing.List[str],
    pp_defines: typing.List[str],
    casters: CastersData,
    dst_dat: typing.Optional[pathlib.Path],
    dst_depfile: typing.Optional[pathlib.Path],
    report_only: bool,
    warn_on_missing_header: bool = True,
):

    try:
        # semiwrap requires user to create yaml files first using update-yaml
        data = AutowrapConfigYaml.from_file(src_yml)
    except FileNotFoundError:
        if not report_only:
            raise

        if warn_on_missing_header:
            print("WARNING: could not find", src_yml)

        data = AutowrapConfigYaml()

    deptarget = None
    if dst_depfile is not None:
        assert dst_dat is not None
        deptarget = [str(dst_dat)]

    if compiler_flavor == "gcc":

        def make_preprocessor(*args, **kwargs):
            return preprocessor.make_gcc_preprocessor(
                print_cmd=False, gcc_args=compiler_args, *args, **kwargs
            )

    # elif compiler_flavor == "msvc":
    #   .. cxxheaderparser's msvc support doesn't generate a depfile, so it's not usable
    #      without breaking incremental build
    else:
        make_preprocessor = preprocessor.make_pcpp_preprocessor

    popts = ParserOptions(
        preprocessor=make_preprocessor(
            defines=pp_defines,
            include_paths=include_paths,
            encoding=data.encoding,
            depfile=dst_depfile,
            deptarget=deptarget,
        )
    )

    gendata = GeneratorData(data, src_yml)

    try:
        hctx = parse_header(
            name,
            src_h,
            src_h_root,
            gendata,
            popts,
            casters,
            report_only,
        )
    except Exception as e:
        raise ValueError(f"processing {src_h}") from e

    missing = gendata.get_missing()

    if not report_only and missing and not data.defaults.ignore:
        print("WARNING: some items not in", src_yml, "for", src_h)
        print(format_missing(missing))

    if dst_dat is not None:
        with open(dst_dat, "wb") as fp:
            pickle.dump(hctx, fp)

    return missing


def make_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-I", "--include-paths", action="append", default=[])
    parser.add_argument("-D", "--defines", action="append", default=[])
    parser.add_argument("--cpp")
    parser.add_argument("name")
    parser.add_argument("src_yml", type=pathlib.Path)
    parser.add_argument("src_h", type=pathlib.Path)
    parser.add_argument("src_h_root", type=pathlib.Path)
    parser.add_argument("in_casters", type=pathlib.Path)
    parser.add_argument("dst_dat", type=pathlib.Path)
    parser.add_argument("dst_depfile", type=pathlib.Path)
    parser.add_argument("compiler_flavor")
    parser.add_argument("cpp_std")
    parser.add_argument("compiler_args", nargs="+")
    parser.add_argument("--update-yaml", action="store_true", default=False)
    return parser


def main():
    parser = make_argparser()
    args = parser.parse_args()

    if not args.update_yaml:
        dst_dat = args.dst_dat
        dst_depfile = args.dst_depfile
        report_only = False
        warn_on_missing_header = True

        with open(args.in_casters, "rb") as fp:
            casters = pickle.load(fp)
    else:
        dst_dat = None
        dst_depfile = None
        report_only = True
        warn_on_missing_header = False
        casters = {}

    compiler_args = args.compiler_args

    if args.compiler_flavor == "gcc":
        compiler_args.append(f"-std={args.cpp_std}")
    else:
        compiler_args.append(f"/std:{args.cpp_std}")

    if args.cpp and args.compiler_flavor != "gcc":
        args.defines.append(f"__cplusplus {args.cpp}")

    missing = generate_wrapper(
        name=args.name,
        src_yml=args.src_yml,
        src_h=args.src_h,
        src_h_root=args.src_h_root,
        dst_dat=dst_dat,
        dst_depfile=dst_depfile,
        include_paths=args.include_paths,
        compiler_flavor=args.compiler_flavor,
        compiler_args=compiler_args,
        casters=casters,
        pp_defines=args.defines,
        report_only=report_only,
        warn_on_missing_header=warn_on_missing_header,
    )

    if args.update_yaml:
        report = format_missing(missing)
        args.src_yml.parent.mkdir(parents=True, exist_ok=True)
        with open(args.src_yml, "w") as fp:
            fp.write(report)


if __name__ == "__main__":
    main()



import subprocess
import sys
import tomli


class BuildDep:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "build-dep", help="Install build dependencies", parents=[parent_parser]
        )
        parser.add_argument("--install", help="Actually do it", action="store_true")
        parser.add_argument(
            "--pre",
            action="store_true",
            default=False,
            help="Include pre-release and development versions.",
        )
        parser.add_argument("--find-links", help="Find links arg", default=None)
        return parser

    def run(self, args):
        with open("pyproject.toml", "rb") as fp:
            pyproject = tomli.load(fp)
        requirements = pyproject.get("build-system", {}).get("requires", [])

        pipargs = [
            sys.executable,
            "-m",
            "pip",
            "--disable-pip-version-check",
            "install",
        ]
        if args.pre:
            pipargs.append("--pre")
        if args.find_links:
            pipargs.extend(
                [
                    "--find-links",
                    args.find_links,
                ]
            )

        pipargs.extend(requirements)
        print(" ".join(pipargs))

        if args.install:
            p = subprocess.run(pipargs)
            return p.returncode



import pathlib
import typing as T

import packaging.markers
import tomli

from .config.util import parse_input
from .config.pyproject_toml import ExtensionModuleConfig, SemiwrapToolConfig


class PyProject:
    def __init__(self, pyproject_path: T.Optional[pathlib.Path] = None) -> None:

        if pyproject_path:
            self.root = pyproject_path.parent.resolve()
        else:
            self.root = pathlib.Path().resolve()

        self._platform = None
        self._project = None
        self._package_root = None

        self._all_deps = None

        self._evaluated_markers: T.Dict[str, bool] = {}

    def _enable_if(self, condition: str) -> bool:
        """
        Evaluates a string containing PEP 508 environment markers
        """
        ok = self._evaluated_markers.get(condition)
        if ok is None:
            ok = packaging.markers.Marker(condition).evaluate()
            self._evaluated_markers[condition] = ok

        return ok

    @property
    def package_root(self) -> pathlib.Path:
        if self._package_root is None:
            # try to detect packages based on the extension modules
            for package_name in self.project.extension_modules.keys():
                parent = pathlib.Path(*package_name.split(".")[:-1])
                if (self.root / parent / "__init__.py").exists():
                    self._package_root = self.root
                    break
                elif (self.root / "src" / parent / "__init__.py").exists():
                    self._package_root = self.root / "src"
                    break
            else:
                raise ValueError("Cannot determine package root")

        return self._package_root

    @property
    def project(self):
        if self._project is None:
            project_fname = self.root / "pyproject.toml"

            try:
                with open(project_fname, "rb") as fp:
                    self.pyproject = tomli.load(fp)
            except FileNotFoundError as e:
                raise ValueError("current directory is not a semiwrap project") from e

            self.project_dict = self.pyproject.get("tool", {}).get("semiwrap", {})

            try:
                self._project = parse_input(
                    self.project_dict, SemiwrapToolConfig, project_fname
                )
            except Exception as e:
                raise ValueError(
                    f"semiwrap configuration in pyproject.toml is incorrect"
                ) from e
        return self._project

    def get_extension(self, module_package_name: str) -> ExtensionModuleConfig:
        return self.project.extension_modules[module_package_name]

    def get_extension_deps(self, extension: ExtensionModuleConfig) -> T.List[str]:
        deps = []
        for wrap in extension.wraps:
            if wrap not in extension.depends:
                deps.append(wrap)
        deps.extend(extension.depends)
        return deps

    def get_extension_yaml_path(self, extension: ExtensionModuleConfig) -> pathlib.Path:
        # Returns relative path

        if extension.yaml_path is None:
            return pathlib.Path("semiwrap")

        yaml_path = pathlib.Path(pathlib.PurePosixPath(extension.yaml_path))
        assert not yaml_path.is_absolute()
        return yaml_path

    def get_extension_headers(
        self, extension: ExtensionModuleConfig
    ) -> T.Generator[T.Tuple[str, str], None, None]:
        for yml, hdr in extension.headers.items():
            if isinstance(hdr, str):
                yield yml, hdr
            elif self._enable_if(hdr.enable_if):
                yield yml, hdr.header



import fnmatch
import glob
import pathlib
from itertools import chain
from os.path import join, relpath
from pathlib import Path, PurePosixPath
import typing as T

from ..pkgconf_cache import PkgconfCache
from ..pyproject import PyProject


class HeaderScanner:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "scan-headers",
            help="Generate a list of wrappable headers in TOML form",
            parents=[parent_parser],
        )
        parser.add_argument("--all", default=False, action="store_true")
        parser.add_argument(
            "--as-ignore",
            default=False,
            action="store_true",
            help="Emit scan_headers_ignore instead",
        )
        parser.add_argument(
            "--check",
            action="store_true",
            help="Exit with error code if any headers printed out",
        )
        parser.add_argument(
            "--pyproject_toml",
            type=pathlib.Path,
            help="The location of the pyproject toml configuration file. By default it will attempt to find it in the current working directory",
            default=pathlib.Path("./pyproject.toml"),
        )
        return parser

    def _make_search_paths(self, pyproject: PyProject) -> T.Dict[str, T.List[Path]]:
        # This logic is copied from makepath, probably should put it somewhere
        # reusable instead?

        # extension: paths
        cache = PkgconfCache()
        search_paths: T.Dict[str, T.List[Path]] = {}

        for name, cfg in pyproject.project.extension_modules.items():
            paths: T.Dict[Path, bool] = {}

            for dep in cfg.wraps:
                entry = cache.get(dep)
                for incpath in entry.include_path:
                    paths[incpath] = True

            for inc in cfg.includes:
                paths[pyproject.root / PurePosixPath(inc)] = True

            path_elems = name.split(".")
            package_path = Path(*path_elems[:-1])

            paths[pyproject.package_root / package_path] = True

            search_paths[name] = list(paths.keys())

        return search_paths

    def run(self, args):
        pyproject = PyProject(args.pyproject_toml)
        project = pyproject.project

        # Get the search path for each extension module
        search_paths = self._make_search_paths(pyproject)

        to_ignore = ["*/trampolines/*", "trampolines/*"] + project.scan_headers_ignore

        def _should_ignore(f):
            for pat in to_ignore:
                if fnmatch.fnmatch(f, pat):
                    return True
            return False

        all_present = set()
        all_missing = set()

        has_difference = False

        if not args.all:
            for ccfg in project.export_type_casters.values():
                incs = [pyproject.root / inc for inc in ccfg.includedir]
                for h in ccfg.headers:
                    for inc in incs:
                        if (inc / h.header).exists():
                            all_present.add(inc / h.header)

            for name, ext in project.extension_modules.items():
                files = []
                for _, f in ext.headers.items():
                    if isinstance(f, str):
                        files.append(Path(f))
                    else:
                        files.append(Path(f.header))

                if not files:
                    continue

                present = set()
                for incdir in search_paths[name]:
                    incdir = Path(incdir)

                    for f in files:
                        if (incdir / f).exists():
                            present.add(f)
                            all_present.add(incdir / f)

                all_missing |= set(files) - present

        all_search_paths = set()
        for ps in search_paths.values():
            for p in ps:
                all_search_paths.add(p)

        for incdir in sorted(all_search_paths, key=lambda pth: -len(pth.parts)):
            files: T.List[Path] = []

            for f in chain(
                glob.glob(join(incdir, "**", "*.h"), recursive=True),
                glob.glob(join(incdir, "**", "*.hpp"), recursive=True),
            ):
                rf = relpath(f, incdir)

                if _should_ignore(rf):
                    all_present.add(incdir / rf)
                    continue

                if (incdir / rf) in all_present:
                    continue

                files.append(Path(rf))

            if not files:
                continue

            has_difference = True
            files.sort()

            if args.as_ignore:
                comment = "    #"
            else:
                comment = "#"

            lastdir = None
            for f in files:
                thisdir = f.parent
                if lastdir is None:
                    if thisdir:
                        print(comment, thisdir)
                elif lastdir != thisdir:
                    print()
                    if thisdir:
                        print(comment, thisdir)
                lastdir = thisdir

                base = f.stem
                if args.as_ignore:
                    print(f'    "{f.as_posix()}",')
                else:
                    print(f'{base} = "{f.as_posix()}"')
            print()

        if all_missing:
            has_difference = True
            print()
            for f in sorted(all_missing):
                print(f"# missing: {f}")

        if args.check:
            return not has_difference



import concurrent.futures
import os
import pathlib
import subprocess
import sys
import traceback
import tempfile
import typing as T
import shutil
import difflib
import dictdiffer
import dictdiffer.utils
from io import StringIO
from ruamel.yaml import YAML

from ..makeplan import InputFile, makeplan, BuildTarget, CompilerInfo
from ..pyproject import PyProject


class YamlUpdater:
    """
    This class will parse the headers for a semiwrap project and create or update their corresponding yaml files.

    The process for doing this goes as such:
        1. Parse the headers and generate .yml files into a temporary directory
        2. Merge the additions / deletions from the fresh generation with the backups
        3. Output the differences
        4. Optionally write the changes to the files

    This process will keep hand edited values like "ignore", "extra_includes", etc, but should be smart enough to delete config
    items that no longer exist.
    """

    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "update-yaml",
            help="Updates YAML files from parsed header files",
            parents=[parent_parser],
        )
        parser.add_argument(
            "--project_file",
            help="The path to the pyproject.toml file",
            type=pathlib.Path,
            default=pathlib.Path("./pyproject.toml"),
        )
        parser.add_argument(
            "--write", help="Write to files if they don't exist", action="store_true"
        )
        parser.add_argument(
            "-v", "--verbose", help="Show full traceback", action="store_true"
        )

        max_jobs = os.cpu_count() or 1

        parser.add_argument(
            "-j",
            "--max-jobs",
            help="Number of processes to run in parallel",
            default=max_jobs,
        )

        return parser

    def run(self, args):
        try:
            with tempfile.TemporaryDirectory() as gendir:
                return self._run(args, pathlib.Path(gendir))
        except Exception as e:
            # Reading the stack trace is annoying, most of the time the exception content
            # is enough to figure out what you did wrong.
            if args.verbose:
                raise

            msg = [
                "ERROR: exception occurred when generating YAML from `pyproject.toml` config\n\n",
            ]

            msg += traceback.format_exception_only(type(e), e)
            cause = e.__context__
            while cause is not None:

                el = traceback.format_exception_only(type(cause), cause)
                el[0] = f"- caused by {el[0]}"
                msg += el

                if cause.__suppress_context__:
                    break

                cause = cause.__context__

            msg.append("\nUse -v/--verbose option for stacktrace")

            print("".join(msg), file=sys.stderr)
            sys.exit(1)

    def _exec_header2dat(self, argv):
        args = [sys.executable, "-m", "semiwrap.cmd.header2dat"] + argv
        proc = subprocess.run(
            args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding="utf-8"
        )
        if proc.returncode == 0:
            return

        return "+" + " ".join(args) + "\n" + proc.stdout.rstrip()

    def _run(self, args, generated_dir: pathlib.Path):
        project_root = args.project_file.parent

        # Problem: if another hatchling plugin sets PKG_CONFIG_PATH to include a .pc
        # file, makeplan() will fail to find it, which prevents a semiwrap program
        # from consuming those .pc files.
        #
        # We search for .pc files in the project root by default and add anything found
        # to the PKG_CONFIG_PATH to allow that to work. Probably won't hurt anything?

        pcpaths: T.Set[str] = set()
        for pcfile in project_root.glob("**/*.pc"):
            pcpaths.add(str(pcfile.parent))

        if pcpaths:
            # Add to PKG_CONFIG_PATH so that it can be resolved by other hatchling
            # plugins if desired
            pkg_config_path = os.environ.get("PKG_CONFIG_PATH")
            if pkg_config_path is not None:
                os.environ["PKG_CONFIG_PATH"] = os.pathsep.join(
                    (pkg_config_path, *pcpaths)
                )
            else:
                os.environ["PKG_CONFIG_PATH"] = os.pathsep.join(pcpaths)

        executor = concurrent.futures.ThreadPoolExecutor(max_workers=args.max_jobs)
        futs = []

        plan = makeplan(project_root, missing_yaml_ok=True)
        for item in plan:
            if not isinstance(item, BuildTarget) or item.command != "header2dat":
                continue

            # convert args to string so we can parse it
            # .. this is weird, but less annoying than other alternatives
            #    that I can think of?
            argv = []
            for arg in item.args:
                if isinstance(arg, str):
                    argv.append(arg)
                elif isinstance(arg, InputFile):
                    # Causes missing content to be written to generated directory
                    if arg.path.name.endswith(".yml"):
                        argv.append(str(generated_dir / arg.path))
                    else:
                        argv.append(str(arg.path))
                elif isinstance(arg, pathlib.Path):
                    argv.append(str(arg.absolute()))
                elif isinstance(arg, CompilerInfo):
                    argv += ["pcpp", "ignored", "ignored"]
                else:
                    # anything else shouldn't matter
                    argv.append("ignored")

            argv.append("--update-yaml")

            # Execute the tool in parallel
            futs.append(executor.submit(self._exec_header2dat, argv))

        fail = False
        for fut in concurrent.futures.as_completed(futs):
            result = fut.result()
            if result is not None:
                fail = True
                print(result)

        if fail:
            return False

        files_updated = self.merge_data(args.write, project_root, generated_dir)

        if args.write:
            print(files_updated, "files were updated")
            return True

        # When not writing, return True if no changes needed, False otherwise

        if files_updated == 0:
            print("All files up to date")
        else:
            print(
                files_updated,
                "files need to be updated (use --write to apply the changes)",
            )

        return files_updated == 0

    def merge_data(
        self,
        write: bool,
        project_root: pathlib.Path,
        generated_directory: pathlib.Path,
    ):
        files_updated = 0

        if not write:
            print("\n\n" + "*" * 20 + "\nDry Run Results\n" + "*" * 20)

        # Collect original YAML files for diff
        original_files = set()
        pyproject = PyProject(project_root)
        for extcfg in pyproject.project.extension_modules.values():
            original_files |= set(
                pyproject.get_extension_yaml_path(extcfg).glob("**/*.yml")
            )

        generated_files = set()
        for f in generated_directory.glob("**/*.yml"):
            generated_files.add(f.relative_to(generated_directory))

        # Delete files that are no longer used in generation
        deleted_files = original_files.difference(generated_files)
        for file_to_delete in deleted_files:
            files_updated += 1
            if write:
                print(f"Deleting unused file {file_to_delete}")
                os.unlink(file_to_delete)
            else:
                print(f"Would delete {file_to_delete}")

        # Add new files
        added_files = generated_files.difference(original_files)
        for f in added_files:
            files_updated += 1
            output_file = project_root / f
            if write:
                output_file.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(generated_directory / f, output_file)
                print(f"Added new file {output_file}")
            else:
                print(f"Would create {output_file}")

            print((generated_directory / f).read_text())

        # Setup the yaml parser with some default arguments
        yaml_ = YAML()
        yaml_.default_flow_style = False
        yaml_.preserve_quotes = True
        yaml_.width = 4096  # Super long width to prevent line wrapping

        GENERATED_KEYWORDS = [
            "enums",
            "functions",
            "classes",
            "methods",
            "attributes",
            "overloads",
        ]

        common_files = original_files.intersection(generated_files)
        for f in sorted(common_files):
            with open(generated_directory / f) as fp:
                generated = yaml_.load(fp)
            with open(project_root / f) as fp:
                original = yaml_.load(fp)

            diffs = dictdiffer.diff(original, generated)

            additions = []

            for diff in diffs:
                action = diff[0]
                # The freshly generated version has added something. We will track a list and apply it at the end, in case that chunk is "ignored"
                if action == "add":
                    additions.append(diff)

                elif action == "change":
                    old_value, new_value = diff[2]
                    # When the new value is None, it is basically equivalent to a deletion, which we don't want to do.
                    if new_value is not None:
                        original = dictdiffer.patch([diff], original)

                # The freshly generated version has removed something. This might be a legitimate removal, like a function being deleted, or it might be deleting some hand tweaked code that we want to keep.
                elif action == "remove":
                    removals = diff[2]
                    for removal in removals:
                        # Make a patch that contains just this one removal
                        modified_diff = [(diff[0], diff[1], [removal])]

                        # Check if the removal is a full deletion of one of the keywords
                        #
                        # i.e. [('remove', 'classes.frc::Encoder', [('enums', {'IndexingType': None})])]
                        # indicates that the enum block was completely removed
                        if removal[0] in GENERATED_KEYWORDS:
                            original = dictdiffer.patch(modified_diff, original)
                            continue

                        # Check if this removal is a sub-change of one of the keywords
                        #
                        # i.e. [('remove', 'classes.frc::Encoder.methods', [('PIDGet', {'rename': 'pidGet'})])]
                        # indicates that the PIDGet function was removed, while other Encoder.methods remain.
                        for item in GENERATED_KEYWORDS:
                            if diff[1].endswith(item):
                                original = dictdiffer.patch(modified_diff, original)
                                break

            # Patch all of the additions while being aware of if the file / class / etc has been marked as "ignore"
            if not original.get("defaults", {}).get("ignore", False):
                for addition in additions:
                    original_contents = dictdiffer.utils.dot_lookup(
                        original, addition[1]
                    )
                    if original_contents.get("ignore", False):
                        continue
                    original = dictdiffer.patch(additions, original)

            # Output a diff
            output_file = project_root / f

            with open(output_file, "r") as of:
                output_lines = of.readlines()

            strbuff = StringIO()
            yaml_.dump(original, strbuff)
            strbuff.seek(0)
            merged_lines = strbuff.readlines()

            differences = difflib.unified_diff(
                output_lines,
                merged_lines,
                fromfile=str(f),
                tofile=str(f),
            )
            differences = list(differences)

            if differences:
                files_updated += 1

                print("Diff for", output_file)
                for difference in differences:
                    print(difference.rstrip())
                print()

                if write:
                    with open(output_file, "w") as fp:
                        strbuff.seek(0)
                        shutil.copyfileobj(strbuff, fp)

        return files_updated



import pathlib
import shlex
import typing as T

import pkgconf


INITPY_VARNAME = "pkgconf_pypi_initpy"


class CacheEntry:

    def __init__(self, name: str) -> None:
        self.name = name
        self.manual = False

    def _get_pkgconf_data(self, *args) -> str:
        r = pkgconf.run_pkgconf(self.name, *args, capture_output=True)
        if r.returncode != 0:
            msg = [f"Package '{self.name}' is not installed"]
            if r.stderr:
                msg.append("")
                msg.append(f"> " + "\n> ".join(r.stderr.decode("utf-8").splitlines()))
            raise RuntimeError("\n".join(msg))

        return r.stdout.decode("utf-8").strip()

    @property
    def requires(self) -> T.List[str]:
        if not hasattr(self, "_requires"):
            raw = self._get_pkgconf_data("--print-requires")
            self._requires = [r for r in raw.split("\n") if r]

        return self._requires

    @property
    def include_path(self) -> T.List[pathlib.Path]:
        """Only the include path for this package"""
        if not hasattr(self, "_include_path"):
            include_path = []
            raw = self._get_pkgconf_data(
                "--cflags-only-I", "--maximum-traverse-depth=1"
            )
            for i in shlex.split(raw):
                assert i.startswith("-I")
                include_path.append(pathlib.Path(i[2:]).absolute())

            self._include_path = include_path

        return self._include_path

    @property
    def full_include_path(self) -> T.List[pathlib.Path]:
        """Include path for this package and requirements"""
        if not hasattr(self, "_full_include_path"):
            full_include_path = []
            raw = self._get_pkgconf_data("--cflags-only-I")
            for i in shlex.split(raw):
                assert i.startswith("-I")
                full_include_path.append(pathlib.Path(i[2:]).absolute())

            self._full_include_path = full_include_path

        return self._full_include_path

    @property
    def type_casters_path(self) -> T.Optional[pathlib.Path]:
        if not hasattr(self, "_type_casters_path"):
            raw = self._get_pkgconf_data("--path")
            pc_path = pathlib.Path(raw)
            type_caster_cfg = pc_path.with_suffix(".pybind11.json")
            if type_caster_cfg.exists():
                self._type_casters_path = type_caster_cfg
            else:
                self._type_casters_path = None

        return self._type_casters_path

    @property
    def libinit_py(self) -> T.Optional[str]:
        if not hasattr(self, "_libinit_py"):
            raw = self._get_pkgconf_data(f"--variable={INITPY_VARNAME}")
            if raw:
                self._libinit_py = raw
            else:
                self._libinit_py = None

        return self._libinit_py


class PkgconfCache:
    def __init__(self) -> None:
        self._cache: T.Dict[str, CacheEntry] = {}

    def add_local(
        self,
        name: str,
        includes: T.List[pathlib.Path],
        requires: T.List[str],
        libinit_py: T.Optional[str] = None,
    ) -> CacheEntry:
        assert name not in self._cache
        entry = CacheEntry(name)
        entry.manual = True
        entry._include_path = [inc.absolute() for inc in includes]
        entry._full_include_path = entry._include_path[:]
        entry._libinit_py = libinit_py
        entry._type_casters_path = None
        entry._requires = requires[:]
        for req in requires:
            dep = self.get(req)
            entry._full_include_path.extend(dep.full_include_path)
        self._cache[name] = entry
        return entry

    def get(self, depname: str) -> CacheEntry:
        entry = self._cache.get(depname)
        if entry is None:
            self._cache[depname] = entry = CacheEntry(depname)
        return entry



from .buffer import RenderBuffer
from .context import (
    HeaderContext,
    ClassContext,
    ClassTemplateData,
    FunctionContext,
    TrampolineData,
)
from .mangle import trampoline_signature

from . import render_pybind11 as rpybind11


def precomma(v: str) -> str:
    return f", {v}" if v else ""


def postcomma(v: str) -> str:
    return f"{v}, " if v else ""


def using_signature(cls: ClassContext, fn: FunctionContext) -> str:
    return f"{cls.full_cpp_name_identifier}_{fn.cpp_name}"


def render_cls_trampoline_hpp(ctx: HeaderContext, cls: ClassContext) -> str:
    """
    Pieces that go into an trampoline file for a class

    - Trampoline base class (if applicable)
    - Template constructors/method fillers (if applicable)
    """

    r = RenderBuffer()
    r.writeln(
        "// This file is autogenerated. DO NOT EDIT\n"
        "\n"
        "#pragma once\n"
        "#include <semiwrap.h>"
    )

    if ctx.extra_includes_first:
        r.writeln("\n// from extra_includes_first")
        for inc in ctx.extra_includes_first:
            r.writeln(f"#include <{inc}>")

    r.writeln("\n// wrapped header")
    r.writeln(f"\n#include <{ctx.rel_fname}>")

    if ctx.extra_includes:
        r.writeln("\n// from extra_includes")
        for inc in ctx.extra_includes:
            r.writeln(f"#include <{inc}>")

    if cls.trampoline is not None:
        _render_cls_trampoline(r, ctx, cls, cls.trampoline)

    if cls.template is not None:
        _render_cls_template_impl(r, ctx, cls, cls.template)

    if not cls.template and not cls.trampoline:
        # TODO: maybe replace this with something that triggers an error
        #       if the trampoline is used? and a better error message
        r.writeln(f'#error "{cls.full_cpp_name} does not have a trampoline"')

    return r.getvalue()


def _render_cls_trampoline(
    r: RenderBuffer, hctx: HeaderContext, cls: ClassContext, trampoline: TrampolineData
):
    """
    Generate trampoline classes to be used for two purposes:

    * Allow python programs to override virtual functions
    * Allow python programs access to protected members

    This trampoline is used from two different places:
    - To generate a trampoline usable by the class itself
    - Generate a trampoline usable by child classes

    Sometimes these are the same trampoline. The exception is when
    a 'final' method is in the base class, then a separate

    Each trampoline type is placed in a different namespace
    to make our life easier.

    Trampoline functions can be disabled via SWGEN_DISABLE_name_[type_type..]
    """

    if cls.template:
        template_argument_list = cls.template.argument_list
        template_parameter_list = cls.template.parameter_list
    else:
        template_argument_list = ""
        template_parameter_list = ""

    # delete specified methods
    if trampoline.methods_to_disable:
        r.writeln()
        for fn in trampoline.methods_to_disable:
            r.writeln(f"#define SWGEN_DISABLE_{ trampoline_signature(fn) }")

    # include override files for each base -- TODO: exclude some bases?
    if cls.bases:
        r.writeln()
        for base in cls.bases:
            r.writeln(f"#include <trampolines/{ base.full_cpp_name_identifier }.hpp>")

    if cls.namespace:
        r.writeln(f"\nnamespace {cls.namespace.strip('::')} {{")

    if hctx.using_declarations:
        r.writeln()
        for decl in hctx.using_declarations:
            r.writeln(f"using {decl.format()};")

    #
    # Each trampoline has a configuration struct.
    #
    # - Stores the base class that the trampoline is wrapping
    # - Provides a mechanism to detect which base class to use when calling an
    #   overloaded virtual function (each class defines the overloads they have,
    #   and so if it's not defined in this config, then it falls back to the
    #   parent configuration)
    #

    r.writeln(
        f"\ntemplate <{postcomma(template_parameter_list)}typename CfgBase = swgen::EmptyTrampolineCfg>"
    )

    if cls.bases:
        r.writeln(f"struct PyTrampolineCfg_{cls.cpp_name} :")

        with r.indent():
            for base in cls.bases:
                r.writeln(
                    f"{base.namespace_}PyTrampolineCfg_{base.cls_name}<{postcomma(base.template_params)}"
                )

            r.writeln("CfgBase")

            for base in cls.bases:
                r.writeln(">")
    else:
        r.writeln(f"struct PyTrampolineCfg_{cls.cpp_name} : CfgBase")

    r.writeln("{")

    with r.indent():
        r.writeln(f"using Base = {cls.full_cpp_name};\n")

        # specify base class to use for each virtual function
        for fn in trampoline.virtual_methods:
            r.writeln(
                f"using override_base_{ trampoline_signature(fn) } = { cls.full_cpp_name };"
            )

    r.writeln("};")

    if cls.bases:
        # To avoid multiple inheritance here, we define a single base with bases that
        # are all template bases..
        #
        # PyTrampolineBase is another trampoline or our base class
        r.writeln()

        r.writeln(
            f"template <typename PyTrampolineBase{precomma(template_parameter_list)}, typename PyTrampolineCfg>"
        )
        r.writeln(f"using PyTrampolineBase_{cls.cpp_name} =")

        for base in cls.bases:
            r.rel_indent(2)
            r.writeln(f"{base.namespace_}PyTrampoline_{base.cls_name}<")

        with r.indent():
            r.writeln("PyTrampolineBase")

            for base in reversed(cls.bases):
                if base.template_params:
                    r.writeln(f", {base.template_params}")
                r.writeln(", PyTrampolineCfg>")
                r.rel_indent(-2)

        r.write_trim(
            f"""
            ;

            template <typename PyTrampolineBase{ precomma(template_parameter_list) }, typename PyTrampolineCfg>
            struct PyTrampoline_{ cls.cpp_name } : PyTrampolineBase_{ cls.cpp_name }<PyTrampolineBase{ precomma(template_argument_list) }, PyTrampolineCfg> {{
              using PyTrampolineBase_{ cls.cpp_name }<PyTrampolineBase{ precomma(template_argument_list) }, PyTrampolineCfg>::PyTrampolineBase_{ cls.cpp_name };
        """
        )

    else:
        r.writeln()
        r.write_trim(
            f"""
            template <typename PyTrampolineBase{ precomma(template_parameter_list) }, typename PyTrampolineCfg>
            struct PyTrampoline_{ cls.cpp_name } : PyTrampolineBase {{
              using PyTrampolineBase::PyTrampolineBase;
        """
        )

    with r.indent():
        for ccls in cls.child_classes:
            if not ccls.template:
                r.writeln(
                    f"using {ccls.cpp_name} [[maybe_unused]] = typename {ccls.full_cpp_name};"
                )

        for enum in cls.enums:
            if enum.cpp_name:
                r.writeln(
                    f"using {enum.cpp_name} [[maybe_unused]] = typename {enum.full_cpp_name};"
                )

        for typealias in cls.user_typealias:
            r.writeln(f"{typealias};")

        for typealias in cls.auto_typealias:
            r.writeln(f"{typealias};")

        if cls.constants:
            r.writeln()
            for name, constant in cls.constants:
                r.writeln(f"static constexpr auto {name} = {constant};")

        #
        # protected constructors -- only used by the direct child
        #

        for fn in trampoline.protected_constructors:
            r.writeln(
                f"\n#ifdef SWGEN_ENABLE_{cls.full_cpp_name_identifier}_PROTECTED_CONSTRUCTORS"
            )
            with r.indent():
                all_decls = ", ".join(p.decl for p in fn.all_params)
                all_names = ", ".join(p.arg_name for p in fn.all_params)
                r.writeln(f"PyTrampoline_{cls.cpp_name}({all_decls}) :")

                if cls.bases:
                    r.writeln(
                        f"  PyTrampolineBase_{cls.cpp_name}<PyTrampolineBase{precomma(trampoline.tmpl_args)}, PyTrampolineCfg>({all_names})"
                    )
                else:
                    r.writeln(f"  PyTrampolineBase({all_names})")

                r.writeln("{}")
            r.writeln("#endif")

        #
        # virtual methods
        #

        for fn in trampoline.virtual_methods:
            _render_cls_trampoline_virtual_method(r, cls, fn)

        #
        # non-virtual protected methods/attributes
        #

        for fn in trampoline.non_virtual_protected_methods:
            r.writeln(f"\n#ifndef SWGEN_DISABLE_{ trampoline_signature(fn) }")

            # hack to ensure we don't do 'using' twice' in the same class, while
            # also ensuring that the overrides can be selectively disabled by
            # child trampoline functions
            with r.indent():
                r.writeln(f"#ifndef SWGEN_UDISABLE_{ using_signature(cls, fn) }")
                with r.indent():
                    r.write_trim(
                        f"""
                        using { cls.full_cpp_name }::{ fn.cpp_name };
                        #define SWGEN_UDISABLE_{ using_signature(cls, fn) }
                    """
                    )
                r.writeln("#endif")
            r.writeln("#endif")

        if cls.protected_properties:
            r.writeln()
            for prop in cls.protected_properties:
                r.writeln(f"using {cls.full_cpp_name}::{prop.cpp_name};")

        if trampoline.inline_code:
            r.writeln()
            r.write_trim(trampoline.inline_code)

    r.writeln("};\n\n")

    if cls.namespace:
        r.writeln(f"}}; // namespace {cls.namespace}")


def _render_cls_trampoline_virtual_method(
    r: RenderBuffer, cls: ClassContext, fn: FunctionContext
):
    r.writeln(f"\n#ifndef SWGEN_DISABLE_{ trampoline_signature(fn) }")
    with r.indent():

        all_decls = ", ".join(p.decl for p in fn.all_params)
        const = " const" if fn.const else ""
        decl = f"{fn.cpp_return_type} {fn.cpp_name}({all_decls}){const}{fn.ref_qualifiers} override {{"
        r.writeln(decl)

        with r.indent():
            if fn.trampoline_cpp_code:
                r.write_trim(fn.trampoline_cpp_code)
            elif fn.ignore_pure:
                r.writeln('throw std::runtime_error("not implemented");')

            else:
                # TODO: probably will break for things like out parameters, etc #}
                if fn.virtual_xform:
                    r.writeln(f"auto custom_fn = {fn.virtual_xform};")

                # We define a "LookupBase" and "CallBase" here because to find the python
                # override we need to use the actual class currently being overridden, but
                # to make the actual call we might need to use a base class.
                #
                # .. lots of duplication here, but it's worse without it

                r.writeln("using LookupBase = typename PyTrampolineCfg::Base;")

                all_names = ", ".join(p.arg_name for p in fn.all_params)
                all_vnames = ", ".join(p.virtual_call_name for p in fn.all_params)

                if fn.is_pure_virtual and fn.virtual_xform:
                    r.write_trim(
                        f"""
                        SEMIWRAP_OVERRIDE_PURE_CUSTOM_NAME({cls.cpp_name}, PYBIND11_TYPE({fn.cpp_return_type}), LookupBase,
                          "{fn.py_name}", {fn.cpp_name}, {all_names});
                        """
                    )
                elif fn.is_pure_virtual:
                    r.write_trim(
                        f"""
                        SEMIWRAP_OVERRIDE_PURE_NAME({cls.cpp_name}, PYBIND11_TYPE({fn.cpp_return_type}), LookupBase,
                          "{fn.py_name}", {fn.cpp_name}, {all_names});
                        """
                    )
                elif fn.virtual_xform:
                    r.write_trim(
                        f"""
                        using CxxCallBase = typename PyTrampolineCfg::override_base_{trampoline_signature(fn)};
                        SEMIWRAP_OVERRIDE_CUSTOM_IMPL(PYBIND11_TYPE({fn.cpp_return_type}), LookupBase,
                          "{fn.py_name}", {fn.cpp_name}, {all_names});
                        return CxxCallBase::{fn.cpp_name}({all_vnames});
                        """
                    )
                else:
                    r.write_trim(
                        f"""
                        using CxxCallBase = typename PyTrampolineCfg::override_base_{trampoline_signature(fn)};
                        PYBIND11_OVERRIDE_IMPL(PYBIND11_TYPE({fn.cpp_return_type}), LookupBase,
                          "{fn.py_name}", {all_names});
                        return CxxCallBase::{fn.cpp_name}({all_vnames});
                        """
                    )

        r.writeln("}")
    r.writeln("#endif")


def _render_cls_template_impl(
    r: RenderBuffer, hctx: HeaderContext, cls: ClassContext, template: ClassTemplateData
):
    if hctx.type_caster_includes:
        r.writeln()
        for inc in hctx.type_caster_includes:
            r.writeln(f"#include <{inc}>")

    r.writeln("\nnamespace swgen {")

    if cls.namespace:
        r.writeln(f"\nusing namespace {cls.namespace};")

    if hctx.using_declarations:
        r.writeln()
        for decl in hctx.using_declarations:
            r.writeln(f"using {decl.format()};")

    r.writeln(f"\ntemplate <{template.parameter_list}>")
    r.writeln(f"struct bind_{cls.full_cpp_name_identifier} {{")

    with r.indent():
        rpybind11.cls_user_using(r, cls)
        rpybind11.cls_auto_using(r, cls)
        rpybind11.cls_consts(r, cls)
        rpybind11.cls_decl(r, cls)

        r.writeln("\npy::module &m;\nstd::string clsName;")

        r.writeln(
            f"bind_{cls.full_cpp_name_identifier}(py::module &m, const char * clsName) :"
        )

        with r.indent():

            # TODO: embedded structs will fail here
            rpybind11.cls_init(r, cls, "clsName")
            r.writeln("m(m),")
            r.writeln("clsName(clsName) {")

            rpybind11.cls_def_enum(r, cls, cls.var_name)

        r.write_trim(
            """
            }
            
            void finish(const char * set_doc = NULL, const char * add_doc = NULL) {
            """
        )
        with r.indent():
            rpybind11.cls_def(r, cls, cls.var_name)

            r.write_trim(
                f"""
                if (set_doc) {{
                  {cls.var_name}.doc() = set_doc;
                }}
                if (add_doc) {{
                  {cls.var_name}.doc() = py::cast<std::string>({cls.var_name}.doc()) + add_doc;
                }}
            """
            )

            if template.inline_code:
                r.writeln()
                r.write_trim(template.inline_code)

        r.writeln("}")

    r.write_trim(
        f"""
        }}; // struct bind_{cls.full_cpp_name_identifier}

        }}; // namespace swgen
        """
    )



"""
Creates an output .pyi file from a given python module.

Arguments are:
    package outpath [subpackage outpath...] -- package mapped_file
"""

import importlib.util
import inspect
import os
from os.path import dirname, join
import pathlib
import shutil
import sys
import tempfile
import typing as T

import pybind11_stubgen


class _PackageFinder:
    """
    Custom loader to allow loading built modules from their location
    in the build directory (as opposed to their install location)
    """

    # Set this to mapping returned from _BuiltEnv.setup_built_env
    mapping: T.Dict[str, str] = {}

    @classmethod
    def find_spec(cls, fullname, path, target=None):
        m = cls.mapping.get(fullname)
        if m:
            return importlib.util.spec_from_file_location(fullname, m)


def _write_pyi(
    package_name, generated_pyi: T.Dict[pathlib.PurePosixPath, pathlib.Path]
):

    # We can't control where stubgen writes files, so tell it to output
    # to a temporary directory and then we copy the files from there to
    # our desired location
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_pth = pathlib.Path(tmpdir)

        # Call pybind11-stubgen
        sys.argv = [
            "<dummy>",
            "--exit-code",
            "--ignore-invalid-expressions=<.*>",
            "--root-suffix=",
            "-o",
            tmpdir,
            package_name,
        ]

        # Create the parent directories in the temporary directory
        for infile in generated_pyi.keys():
            (tmpdir_pth / infile).parent.mkdir(parents=True, exist_ok=True)

        pybind11_stubgen.main()

        # stubgen doesn't take a direct output filename, so move the file
        # to our desired location
        for infile, output in generated_pyi.items():
            output.unlink(missing_ok=True)
            shutil.move(tmpdir_pth / infile, output)


def main():

    generated_pyi: T.Dict[pathlib.PurePosixPath, pathlib.Path] = {}
    argv = sys.argv

    if len(argv) < 3:
        print(inspect.cleandoc(__doc__ or ""), file=sys.stderr)
        sys.exit(1)

    # Package name first
    package_name = argv[1]

    # Output file map: input output
    idx = 2
    while idx < len(argv):
        if argv[idx] == "--":
            idx += 1
            break

        generated_pyi[pathlib.PurePosixPath(argv[idx])] = pathlib.Path(argv[idx + 1])
        idx += 2

    # Arguments are used to set up the package map
    package_map = _PackageFinder.mapping
    for i in range(idx, len(argv), 2):
        # python 3.9 requires paths to be resolved
        package_map[argv[i]] = os.fspath(pathlib.Path(argv[i + 1]).resolve())

    # Add parent packages too
    # .. assuming there are __init__.py in each package
    for pkg in list(package_map.keys()):
        while True:
            idx = pkg.rfind(".")
            if idx == -1:
                break
            ppkg = pkg[:idx]
            if ppkg not in package_map:
                package_map[ppkg] = join(
                    dirname(dirname(package_map[pkg])), "__init__.py"
                )
            pkg = ppkg

    sys.meta_path.insert(0, _PackageFinder)

    _write_pyi(package_name, generated_pyi)


if __name__ == "__main__":
    main()



import dataclasses
import os
import pathlib
import typing as T


def _escape_dep(dep: str):
    dep = dep.replace("\\", "\\\\")
    dep = dep.replace(" ", "\\ ")
    return dep


@dataclasses.dataclass
class Depfile:
    # TODO: currently only supports single output target

    target: pathlib.Path
    deps: T.List[pathlib.Path] = dataclasses.field(default_factory=list)

    def add(self, dep: pathlib.Path):
        self.deps.append(dep)

    def write(self, path: pathlib.Path):
        """
        Write make-compatible depfile
        """
        with open(path, "w") as fp:
            target = _escape_dep(str(self.target.absolute()))
            fp.write(f"{target}:")
            for dep in self.deps:
                fp.write(f" \\\n  {_escape_dep(str(dep.absolute()))}")
            fp.write("\n")



from .buffer import RenderBuffer
from .context import HeaderContext


def render_class_prologue(r: RenderBuffer, hctx: HeaderContext):
    # fmt: off
    r.writeln(
        "// This file is autogenerated. DO NOT EDIT\n"
        "#include <semiwrap.h>"
    )
    # fmt: on

    if hctx.has_vcheck:
        r.writeln("#include <functional>")

    if hctx.extra_includes_first:
        r.writeln()
        for inc in hctx.extra_includes_first:
            r.writeln(f"#include <{inc}>")

    r.writeln(f"\n#include <{hctx.rel_fname}>")

    if hctx.type_caster_includes:
        r.writeln()
        for inc in hctx.type_caster_includes:
            r.writeln(f"#include <{inc}>")

    if hctx.need_operators_h:
        r.writeln(f"\n#include <pybind11/operators.h>")

    if hctx.using_declarations:
        r.writeln()
        for decl in hctx.using_declarations:
            r.writeln(f"using {decl.format()};")

    for cls in hctx.classes_with_trampolines:
        r.writeln()
        r.writeln(
            f"#define SWGEN_ENABLE_{cls.full_cpp_name_identifier}_PROTECTED_CONSTRUCTORS"
        )
        r.writeln(f"#include <trampolines/{cls.full_cpp_name_identifier}.hpp>")



#
# pyproject.toml
#

import dataclasses
import re
from typing import Dict, List, Optional, Union

_arch_re = re.compile(r"\{\{\s*ARCH\s*\}\}")
_os_re = re.compile(r"\{\{\s*OS\s*\}\}")


@dataclasses.dataclass
class TypeCasterHeader:

    #: Header file to include when one of the types are detected in a wrapper
    header: str

    #: Types to look for to indicate that this type caster header should be
    #: included.
    types: List[str]

    #: If a parameter type that requires this type caster requires a default
    #: argument, a C-style ``(type)`` cast is used on the default argument.
    #:
    #: The default cast can be disabled via param_override's ``disable_type_caster_default_cast``
    default_arg_cast: bool = False


@dataclasses.dataclass
class TypeCasterConfig:
    """
    Specifies type casters to be exported as a pkgconf package. When a semiwrap package has a
    dependency on such a package, it will use this type information at generation time and
    include the type caster headers in generated wrappers.

    .. code-block:: toml

       [tool.semiwrap.export_type_casters.NAME]
       pypackage = "mypkg"
       includedirs = ["src/mypkg/type-casters"]

       [[tool.semiwrap.export_type_casters.NAME.headers]]
       header = "my_type_caster.h"
       types = ["foo_t", "ns::ins::bar_t"]

    .. seealso:: :ref:`type_casters`
    """

    #: Python package to write the .pc and .json file to
    pypackage: str

    #: Location of type caster header files, relative to the root of the project
    includedir: List[str]

    #: pkgconf 'requires'
    requires: List[str] = dataclasses.field(default_factory=list)

    #: Contains header/type mappings for type casters
    headers: List[TypeCasterHeader] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class ConditionalHeader:
    """
    Allows specifying that a header will only be autogenerated if the specified
    condition is true.

    .. code-block:: toml

        [tool.semiwrap.extension_modules."PACKAGE.NAME".headers]
        Name = { header="header.h", enable_if = "platform_machine == 'aarch64'" }

    """

    #: Name of the header file
    header: str

    #: This is a PEP 508 environment marker specification.
    #:
    #: The header will not be parsed if this does not evaluate to true
    enable_if: str


@dataclasses.dataclass
class ExtensionModuleConfig:
    """
    Configuration for generating pybind11 wrappers around specified header files.

    .. code-block:: toml

        [tool.semiwrap.extension_modules."PACKAGE.NAME"]

    """

    #: The name used internally to refer to this extension module. The name is
    #: used as the prefix of variable names in build files.
    #:
    #: If not specified, the default name is the package name with invalid
    #: characters replaced by underscores.
    name: Optional[str] = None

    #: Name of generated file that ensures the shared libraries and any
    #: dependencies are loaded. Defaults to ``_init_XXX.py``, where XXX
    #: is the last element of the package name
    #:
    #: Generally, you should create an ``__init__.py`` file that imports
    #: this module, otherwise your users will need to do so.
    libinit: Optional[str] = None

    #: Name of pkgconf libraries that this is wrapping. This is also added to
    #: depends list.
    wraps: List[str] = dataclasses.field(default_factory=list)

    #: List of pkgconf library dependencies. This affects this wrapper
    #: library in the following ways:
    #:
    #: * Any include file directories exported by the dependency will be added
    #:   to the include path for any source files compiled by this wrapper
    #: * It will be linked to any libraries the dependency contains
    #: * The python module for the dependency will be imported in the
    #:   ``_init{extension}.py`` file.
    depends: List[str] = dataclasses.field(default_factory=list)

    #: List of include directories for this module relative to the project root
    includes: List[str] = dataclasses.field(default_factory=list)

    #: #define macros to apply to the compilation
    #:
    #: .. code-block:: toml
    #:
    #:    [tool.semiwrap.extension_modules."package.name".defines]
    #:    SOMETHING_DEFINED = 1
    #:
    defines: Dict[str, Union[str, int]] = dataclasses.field(default_factory=dict)

    #: Specifies header files that autogenerated pybind11 wrappers will be
    #: created for. Simple C++ headers will most likely 'just work', but
    #: complex headers will need to have an accompanying :attr:`generation_data`
    #: file specified that can customize the autogenerated files.
    #:
    #: List of dictionaries: each dictionary key is used for the function
    #: name of the initialization function, the value is the header that is
    #: being wrapped. The header is searched for in the following order:
    #:
    #: - Relative to include directories specified by 'wraps' dependencies
    #: - Relative to the extra include directories for this module
    #: - Relative to the package directory
    #:
    #: .. code-block:: toml
    #:
    #:    [tool.semiwrap.extension_modules."PACKAGE.NAME".headers]
    #:    Name = "header.h"
    #:
    #: .. seealso:: :ref:`autowrap`
    #:
    headers: Dict[str, Union[str, ConditionalHeader]] = dataclasses.field(
        default_factory=dict
    )

    #: Path to a directory of yaml files. Generation data will be looked up
    #: using the key in the headers dictionary.
    #:
    #: These YAML files can be generated via the semiwrap command line tool:
    #:
    #: .. code-block:: sh
    #:
    #:    semiwrap update-yaml --write
    #:
    #: .. seealso:: :ref:`gendata`
    #:
    yaml_path: Optional[str] = None

    #: If True, skip this wrapper
    ignore: bool = False


@dataclasses.dataclass
class SemiwrapToolConfig:
    """
    Contains information for configuring the project

    .. code-block:: toml

       [tool.semiwrap]

    .. note:: This section is required
    """

    #: List of headers for the scan-headers tool to ignore
    scan_headers_ignore: List[str] = dataclasses.field(default_factory=list)

    #: List of python packages with __init__.py to update when ``python setup.py update_init``
    #: is called -- this is an argument to the ``semiwrap create-imports`` command, and
    #: may contain a space and the second argument to create-imports.
    update_init: List[str] = dataclasses.field(default_factory=list)

    #: Modules to autogenerate. The key is the name of the python package that will
    #: be generated. For example, "package.name" will cause the shared library
    #: 'package/name-{abi tag}.so' to be installed
    extension_modules: Dict[str, ExtensionModuleConfig] = dataclasses.field(
        default_factory=dict
    )

    #: Type casters to export. The key is the pkgconf package name.
    export_type_casters: Dict[str, TypeCasterConfig] = dataclasses.field(
        default_factory=dict
    )


@dataclasses.dataclass
class SemiwrapHatchlingConfig:
    """
    semiwrap hatchling hook configuration

    .. code-block:: toml

       [tool.hatch.build.hooks.semiwrap]

    .. note:: This section is required
    """

    #: Output directory to write meson.build to
    autogen_build_path: str = "semiwrap"

    #: Output directory to write second stage meson.build to (defaults
    #: to ``autogen_build_path / 'modules'``)
    module_build_path: Optional[str] = None



"""
Creates a template instance .cpp file from a .dat file created by parsing a header
"""

import inspect
import pathlib
import pickle
import sys

from ..autowrap.context import HeaderContext
from ..autowrap.render_tmpl_inst import render_template_inst_cpp


def _write_wrapper_cpp(input_dat: pathlib.Path, py_name: str, output_cpp: pathlib.Path):
    with open(input_dat, "rb") as fp:
        hctx = pickle.load(fp)

    assert isinstance(hctx, HeaderContext)

    for tmpl in hctx.template_instances:
        if tmpl.py_name == py_name:
            break
    else:
        raise ValueError(f"internal error: cannot find {py_name} in {hctx.orig_yaml}")

    content = render_template_inst_cpp(hctx, tmpl)
    output_cpp.write_text(content, encoding="utf-8")


def main():
    try:
        _, input_dat, py_name, output_cpp = sys.argv
    except ValueError:
        print(inspect.cleandoc(__doc__ or ""), file=sys.stderr)
        sys.exit(1)

    _write_wrapper_cpp(pathlib.Path(input_dat), py_name, pathlib.Path(output_cpp))


if __name__ == "__main__":
    main()



import typing

from validobj import errors
import validobj.validation

T = typing.TypeVar("T")


class ValidationError(Exception):
    pass


def _convert_validation_error(fname, ve: errors.ValidationError) -> ValidationError:
    locs = []
    msg = []

    e = ve
    while e is not None:

        if isinstance(e, errors.WrongFieldError):
            locs.append(f".{e.wrong_field}")
        elif isinstance(e, errors.WrongListItemError):
            locs.append(f"[{e.wrong_index}]")
        else:
            msg.append(str(e))

        e = e.__cause__

    loc = "".join(locs)
    if loc.startswith("."):
        loc = loc[1:]
    msg = "\n  ".join(msg)
    vmsg = f"{fname}: {loc}:\n  {msg}"
    return ValidationError(vmsg)


def parse_input(value: typing.Any, spec: typing.Type[T], fname) -> T:
    try:
        return validobj.validation.parse_input(value, spec)
    except errors.ValidationError as ve:
        raise _convert_validation_error(fname, ve) from None


# yaml converts empty values to None, but we never want that
def fix_yaml_dict(a: typing.Any):
    if isinstance(a, dict):
        for k, v in a.items():
            if v is None:
                a[k] = {}
            if isinstance(v, dict):
                fix_yaml_dict(v)
    elif isinstance(a, list):
        for v in a:
            if isinstance(v, dict):
                fix_yaml_dict(v)

    return a



import inspect
import posixpath
import subprocess
import sys
import typing
import types
import pathlib

from ..pyproject import PyProject


class ImportCreator:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "create-imports",
            help="Generate suitable imports for a module. Prefer using update-init",
            parents=[parent_parser],
        )
        parser.add_argument("base", help="Ex: wpiutil")
        parser.add_argument("compiled", nargs="?", help="Ex: wpiutil._impl.wpiutil")
        parser.add_argument(
            "--write", "-w", action="store_true", help="Modify existing __init__.py"
        )
        parser.add_argument(
            "--override_output_file",
            type=pathlib.Path,
            help="If specified, will write the updated file to this location instead of making the change in place. Only used if --write is set.",
        )
        return parser

    def _rel(self, base: str, compiled: str) -> str:
        base = posixpath.join(*base.split("."))
        compiled = posixpath.join(*compiled.split("."))
        elems = posixpath.relpath(compiled, base).split("/")
        elems = ["" if e == ".." else e for e in elems]
        return f".{'.'.join(elems)}"

    def run(self, args):
        self.create(args.base, args.compiled, args.write, args.override_output_file)

    def create(
        self,
        base: str,
        compiled: typing.Optional[str],
        write: bool,
        override_output_file: typing.Optional[pathlib.Path],
    ):
        # Runtime Dependency Check
        try:
            import black
        except:
            print(
                "Error, The following module is required to run this tool: black",
                file=sys.stderr,
            )
            return 1

        if not compiled:
            compiled = f"{base}._{base.split('.')[-1]}"
            print(f"CHK base={base} compiled={compiled} (auto)")
        else:
            print(f"CHK base={base} compiled={compiled}")

        # TODO: could probably generate this from parsed code, but seems hard
        ctx = {}
        exec(f"from {compiled} import *", {}, ctx)
        for k in list(ctx.keys()):
            if isinstance(ctx[k], types.ModuleType):
                del ctx[k]

        relimport = self._rel(base, compiled)

        stmt_compiled = "" if not compiled else f" {compiled}"
        begin_stmt = f"# autogenerated by 'semiwrap create-imports {base}"
        old_begin_stmt = f"# autogenerated by 'robotpy-build create-imports {base}"

        stmt = inspect.cleandoc(
            f"""

            {begin_stmt}{stmt_compiled}'
            from {relimport} import {','.join(sorted(ctx.keys()))}
            __all__ = ["{'", "'.join(sorted(ctx.keys()))}"]
        
        """
        )

        content = subprocess.check_output(
            ["black", "-", "-q"], input=stmt.encode("utf-8")
        ).decode("utf-8")

        if write:
            fctx = {}
            exec(f"from {base} import __file__", {}, fctx)
            fname = fctx["__file__"]

            with open(fname) as fp:
                fcontent = orig_content = fp.read()

            # Find the beginning statement
            idx = startidx = fcontent.find(begin_stmt)
            if startidx == -1:
                idx = startidx = fcontent.find(old_begin_stmt)

            if startidx != -1:
                for to_find in ("from", "__all__", "[", "]", "\n"):
                    idx = fcontent.find(to_find, idx)
                    if idx == -1:
                        startidx = -1
                        break

            if startidx == -1:
                # If not present, just append and let the user figure it out
                fcontent = fcontent + "\n" + content
            else:
                fcontent = fcontent[:startidx] + content + fcontent[idx + 1 :]

            # Always write the file if override_output_file is specified.
            # Othewise only write if the contents have changed.
            if override_output_file or fcontent != orig_content:
                fname = override_output_file or fname
                with open(fname, "w") as fp:
                    fp.write(fcontent)
                print("MOD", base)
            else:
                print("OK", base)

        else:
            print(content)


class UpdateInit:
    @classmethod
    def add_subparser(cls, parent_parser, subparsers):
        parser = subparsers.add_parser(
            "update-init",
            help="Updates __init__.py files using settings from tool.semiwrap.update_init",
            parents=[parent_parser],
        )
        return parser

    def run(self, args):
        project = PyProject().project

        if project.update_init is None:
            print("[tool.semiwrap].update_init not set", file=sys.stderr)
            return 1

        ic = ImportCreator()

        for to_update in project.update_init:
            if " " in to_update:
                base, compiled = to_update.split(" ", 1)
            else:
                base = to_update
                compiled = None

            ic.create(base, compiled, True)



"""
Generates a pkg-config file for this extension
"""

import argparse
import inspect
import pathlib

from ..mkpc import make_pc_file
from ..pyproject import PyProject


def main():
    parser = argparse.ArgumentParser(usage=inspect.cleandoc(__doc__ or ""))
    parser.add_argument("module_package_name")
    parser.add_argument("name")
    parser.add_argument("pyproject_toml", type=pathlib.Path)
    parser.add_argument("pcfile", type=pathlib.Path)
    parser.add_argument("--libinit-py")
    args = parser.parse_args()

    module_package_name = args.module_package_name
    project = PyProject(args.pyproject_toml)

    module = project.get_extension(module_package_name)
    depends = project.get_extension_deps(module)

    pc_install_path = project.package_root / pathlib.Path(
        *module_package_name.split(".")[:-1]
    )
    make_pc_file(
        project_root=project.root,
        pcfile=args.pcfile,
        pc_install_path=pc_install_path,
        name=args.name,
        desc="semiwrap pybind11 module",
        version="",
        includes=module.includes,
        depends=depends,
        libinit_py=args.libinit_py,
        generator_name="semiwrap.cmd_gen_pkgconf",
    )


if __name__ == "__main__":
    main()



"""
Determines which casters are in the pyproject.toml and publishes them so they
can be consumed and used by other wheels.

Generates a FOO.pc file and a FOO.pybind11.json file
"""

import pathlib
import sys

from ..casters import (
    TypeCasterJsonData,
    TypeCasterJsonHeader,
    save_typecaster_json_data,
)
from ..mkpc import make_pc_file
from ..pyproject import PyProject


def main():
    try:
        _, pyproject_toml, caster_name, output_json, output_pc = sys.argv
    except ValueError:
        print(__doc__, file=sys.stderr)
        sys.exit(1)

    project = PyProject(pathlib.Path(pyproject_toml))
    cfg = project.project.export_type_casters[caster_name]

    # make sure the include directories actually exist
    include_dirs = []
    for inc in cfg.includedir:
        includedir = project.root / pathlib.Path(inc)
        include_dirs.append(includedir)
        if not includedir.exists():
            print(f"ERROR: {includedir} does not exist", file=sys.stderr)
            print(
                f"- specified at [tool.semiwrap.export_type_casters.{caster_name}].includedir",
                file=sys.stderr,
            )
            sys.exit(1)

    pc_install_path = project.package_root / pathlib.Path(*cfg.pypackage.split("."))
    make_pc_file(
        project_root=project.root,
        pcfile=pathlib.Path(output_pc),
        pc_install_path=pc_install_path,
        name=caster_name,
        desc="pybind11 type casters",
        version="",
        includes=cfg.includedir,
        depends=cfg.requires,
        libinit_py=None,
        generator_name="semiwrap.cmd_publish_casters",
    )

    #
    # Gather the data and write it next to the pc file
    #

    data = TypeCasterJsonData()
    for hdr in cfg.headers:

        # Ensure the published header actually exists
        searched = []
        for inc in include_dirs:
            full_hdr = inc / hdr.header
            if full_hdr.exists():
                break

            searched.append(full_hdr)
        else:

            print(f"ERROR: {hdr.header} does not exist", file=sys.stderr)
            print(
                f"- specified at [[tool.semiwrap.export_type_casters.{caster_name}.headers]].header",
                file=sys.stderr,
            )
            for s in searched:
                print(f"- searched '{s}'")
            sys.exit(1)

        data.headers.append(
            TypeCasterJsonHeader(
                header=hdr.header,
                types=hdr.types,
                default_arg_cast=hdr.default_arg_cast,
            )
        )

    save_typecaster_json_data(pathlib.Path(output_json), data)


if __name__ == "__main__":
    main()



from .buffer import RenderBuffer
from .context import HeaderContext

from . import render_pybind11 as rpybind11
from .render_cls_prologue import render_class_prologue


def render_wrapped_cpp(hctx: HeaderContext) -> str:
    """
    This contains the primary binding code generated from parsing a single
    header file. There are also per-class headers generated (templates,
    trampolines), and those are included/used by this.
    """
    r = RenderBuffer()

    render_class_prologue(r, hctx)

    if hctx.template_instances:
        r.writeln(f'\n#include "{hctx.hname}_tmpl.hpp"')

    if hctx.extra_includes:
        r.writeln()
        for inc in hctx.extra_includes:
            r.writeln(f"#include <{inc}>")

    if hctx.user_typealias:
        r.writeln()
        for typealias in hctx.user_typealias:
            r.writeln(f"{typealias};")

    #
    # Ordering of the initialization function
    #
    # - namespace/typealiases
    # - global enums
    # - templates (because CRTP)
    # - class declarations
    # - class enums
    # - class methods
    # - global methods
    #
    # Additionally, we use two-part initialization to ensure that documentation
    # strings are generated properly. First part is to register the class with
    # pybind11, second part is to generate all the methods/etc for it.
    #
    # TODO: make type_traits optional by detecting trampoline

    r.writeln("\n#include <type_traits>")

    if hctx.namespaces:
        r.writeln()
        for ns in hctx.namespaces:
            r.writeln(f"using namespace {ns};")

    r.writeln(f"\nstruct semiwrap_{hctx.hname}_initializer {{\n")

    with r.indent():
        for cls in hctx.classes:
            if not cls.template:
                rpybind11.cls_user_using(r, cls)
                rpybind11.cls_consts(r, cls)

        if hctx.subpackages:
            r.writeln()
            for vname in hctx.subpackages.values():
                r.writeln(f"py::module {vname};")

        # enums
        for index, enum in enumerate(hctx.enums, start=1):
            rpybind11.enum_decl(r, enum, f"enum{index}")

        # template decls
        for tmpl_data in hctx.template_instances:
            if not tmpl_data.matched:
                r.writeln(f"swgen::{tmpl_data.binder_typename} {tmpl_data.var_name};")

        # class decls
        for cls in hctx.classes:
            if cls.template is None:
                r.writeln()
                rpybind11.cls_decl(r, cls)
            elif cls.template.instances:
                r.writeln()
                for tmpl_data in cls.template.instances:
                    r.writeln(
                        f"swgen::{tmpl_data.binder_typename} {tmpl_data.var_name};"
                    )

        r.writeln("\npy::module &m;\n")
        r.writeln(f"semiwrap_{hctx.hname}_initializer(py::module &m) :")

        with r.indent():
            for pkg, vname in hctx.subpackages.items():
                r.writeln(f'{vname}(m.def_submodule("{pkg}")),')

            for index, enum in enumerate(hctx.enums, start=1):
                r.writeln(
                    f"enum{index}({rpybind11.enum_init_args(enum.scope_var, enum)}),"
                )

            for tmpl_data in hctx.template_instances:
                if not tmpl_data.matched:
                    r.writeln(
                        f'{tmpl_data.var_name}({tmpl_data.scope_var}, "{tmpl_data.py_name}"),'
                    )

            for cls in hctx.classes:
                if not cls.template:
                    rpybind11.cls_init(r, cls, f'"{cls.py_name}"')
                else:
                    for tmpl_data in cls.template.instances:
                        r.writeln(
                            f'{tmpl_data.var_name}({tmpl_data.scope_var}, "{tmpl_data.py_name}"),'
                        )

            r.writeln("m(m)")

        if hctx.enums or hctx.classes:
            r.writeln("{")
            with r.indent():
                # enums can go in the initializer because they cant have dependencies,
                # and then we dont need to figure out class dependencies for enum arguments

                for index, enum in enumerate(hctx.enums, start=1):
                    r.writeln(f"enum{index}")
                    with r.indent():
                        rpybind11.enum_def(r, enum.scope_var, enum)

                for cls in hctx.classes:
                    rpybind11.cls_def_enum(r, cls, cls.var_name)
                    for ccls in cls.child_classes:
                        rpybind11.cls_def_enum(r, ccls, ccls.var_name)
            r.writeln("}")
        else:
            r.writeln("{}")

        r.writeln("\nvoid finish() {\n")

        with r.indent():

            # Templates
            for tdata in hctx.template_instances:
                r.writeln(f"\n{tdata.var_name}.finish(")
                with r.indent():
                    if tdata.doc_set:
                        r.writeln(f'{rpybind11.mkdoc("", tdata.doc_set, "")},')
                    else:
                        r.writeln("nullptr,")

                    if tdata.doc_add:
                        r.writeln(rpybind11.mkdoc("", tdata.doc_add, ""))
                    else:
                        r.writeln("nullptr")
                r.writeln(");")

            # Class methods
            for cls in hctx.classes:
                if not cls.template:
                    r.writeln("{")
                    with r.indent():
                        rpybind11.cls_auto_using(r, cls)
                        rpybind11.cls_def(r, cls, cls.var_name)
                    r.writeln("}")

            # Global methods
            if hctx.functions:
                r.writeln()
                for fn in hctx.functions:
                    if not fn.ignore_py:
                        r.writeln(fn.scope_var)
                        with r.indent(1):
                            rpybind11.genmethod(r, None, fn, None)
                        r.writeln(";")

            if hctx.inline_code:
                r.writeln()
                r.write_trim(hctx.inline_code)

        r.writeln("}")

    r.writeln(
        f"}}; // struct semiwrap_{hctx.hname}_initializer\n"
        "\n"
        f"static std::unique_ptr<semiwrap_{hctx.hname}_initializer> cls;\n"
        "\n"
        f"void begin_init_{hctx.hname}(py::module &m) {{\n"
        f"  cls = std::make_unique<semiwrap_{hctx.hname}_initializer>(m);\n"
        "}\n"
        "\n"
        f"void finish_init_{hctx.hname}() {{\n"
        "  cls->finish();\n"
        "  cls.reset();\n"
        "}\n"
    )

    return r.getvalue()
